import os

import numpy as np
import pandas as pd
import numpy as np
import xarray as xr
from datetime import datetime
import matplotlib as mpl
import matplotlib.pyplot as plt
from loguru import logger

from gfatpy.cloudnet import CLOUDNET_INFO
from gfatpy.utils import utils


def get_cloudnet_fn(
    station_name: str, date_str: str, kind: str, cloudnet_dn: str | None = None
) -> str | None:
    """Get file name full path for a day cloudnet data

    Args:
        station_name: str
            name of station
        date_str: str
            date (day) YYYYMMDD
        kind: str
            categorize, classification
        cloudnet_dn: str
            absolute path for cloudnet data
    Returns:
        cldnt_fn: str
            full path cloudnet file name
    """

    try:
        if cloudnet_dn is None:  # GFAT NAS
            cloudnet_dn = r"/mnt/NASGFAT/datos/CLOUDNET"

        if kind in CLOUDNET_INFO["CLOUDNET_KINDS"]:
            date_year = datetime.strptime(date_str, "%Y%m%d").strftime("%Y")
            cldnt_fn = os.path.join(
                cloudnet_dn,
                station_name,
                kind,
                date_year,
                "%s_%s_%s.nc" % (date_str, station_name, kind),
            )
            if not os.path.isfile(cldnt_fn):
                logger.error("File %s not found" % cldnt_fn)
                cldnt_fn = None
        else:
            logger.error("Cloudnet File kind %s does not exist" % kind)
            cldnt_fn = None
    except Exception as e:
        logger.error(str(e))
        cldnt_fn = None

    return cldnt_fn


def CBH_attenuated(
    cat: xr.Dataset,
    threshold: float = 2.5e-6,
    cloud_height_maximum: float = 4000,
    plot_flag: bool = False,
) -> xr.Dataset:

    """
    CBH_attenuated finds the CBH from attenuated backscatter in CLUODNET categorize files
    Input:
    cat: xarray from  CLUODNET categorize file
    threshold: minimum value of cloud attenuated backscatter (default: 2.5e-6 m^1*sr^-1
    cloud_height_maximum: maximxum height at which CBH will be search (Default: 4000 m),
    plot_flag: it plots the CBH temporal evolution (Default: False)
    Output:
    cat: xarray categorize input but with CBH.
    """

    top_idx, _ = utils.find_nearest_1d(cat.height.values, cloud_height_maximum)
    height = cat.height[0:top_idx].values

    CBH = np.nan * np.ones(len(cat.time))
    for i_ in np.arange(len(cat.time)):
        profile = cat.beta[i_, 0:top_idx].values
        try:
            candidate = height[profile > threshold][0]
        except:
            candidate = height[profile > threshold]
        if candidate < cloud_height_maximum:
            CBH[i_] = candidate
    if plot_flag:
        fig, axes = plt.subplots(1, 1)
        for i_ in np.arange(len(cat.time)):
            profile = cat.beta[i_, 0:top_idx].values
            axes.plot(profile, height)
        fig.set_figheight(15)
        fig.set_figwidth(15)
        axes.vlines(threshold, 0, cloud_height_maximum)
        plt.xlim(0, threshold * 3)

    cat["CBH_attenuated_beta"] = ("time", CBH)

    return cat


def cloud_edges(cat: xr.Dataset, threshold_split: float = 200) -> xr.Dataset:
    """
    CTH_finder finds the CTH from radar reflectivity in CLUODNET categorize files
    Input:
    cat: xarray from CLUODNET categorize file
    cloud_height_maximum: maximxum height at which CTH will be search (Default: 5000 m),
    Output:
    cat: xarray categorize input but with CTH.
    """
    cat["Zboolean"] = cat.Z.where(np.isnan(cat.Z), other=1)
    cat["Zboolean"] = cat.Zboolean.where(~np.isnan(cat.Z), other=0)
    cat["Zbooleandiff_CBH"] = (
        ("time", "height"),
        np.diff(cat.Zboolean, axis=1, prepend=0),
    )
    cat["Zbooleandiff_CTH"] = (
        ("time", "height"),
        np.diff(cat.Zboolean, axis=1, prepend=0),
    )

    # CBH candidates
    final_CBH = np.nan * np.ones(np.shape(cat["Z"]))
    final_CTH = np.nan * np.ones(np.shape(cat["Z"]))
    CBH0 = np.nan * np.ones(len(cat["time"]))
    CTH0 = np.nan * np.ones(len(cat["time"]))
    single_layer = np.ones(len(cat["time"]))
    idx_height = np.arange(len(cat["height"]))
    for idx_ in np.arange(len(cat["time"])):
        # CBH candidates
        booleanCBH = cat["Zbooleandiff_CBH"][idx_, :] == 1
        CBHcandidates = cat["height"].where(booleanCBH)
        # CTH candidates
        booleanCTH = cat["Zbooleandiff_CTH"][idx_, :] == -1
        CTHcandidates = cat["height"].where(booleanCTH)

        # Glue clouds too near
        CBHs = CBHcandidates[booleanCBH]
        CTHs = CTHcandidates[booleanCTH]
        if len(CBHs) > 0:
            if np.logical_and.reduce((len(CBHs) > 1, len(CBHs) == len(CTHs))):
                distance = CBHs.values[1:] - CTHs.values[:-1]  # type: ignore
                if np.logical_and.reduce((len(distance) > 0, distance.any())):
                    idx_badDistance = (
                        distance < threshold_split
                    )  # CTH(n) and CBH(n+1) should be far enough to be consider different cloud
                    goodCBH = CBHs[np.append(True, ~idx_badDistance)]
                    goodCTH = CTHs[np.append(~idx_badDistance, True)]
                    if np.logical_and.reduce((goodCBH.any(), goodCTH.any())):
                        for CBH_ in goodCBH:
                            final_CBH[idx_, idx_height[cat["height"] == CBH_]] = CBH_
                        for CTH_ in goodCTH:
                            final_CTH[idx_, idx_height[cat["height"] == CTH_]] = CTH_
                    CBH0[idx_] = final_CBH[idx_, :][~np.isnan(final_CBH[idx_, :])][0]
                    CTH0[idx_] = final_CTH[idx_, :][~np.isnan(final_CTH[idx_, :])][0]
                single_layer[idx_] = 0
            else:
                try:
                    CBH0[idx_] = CBHs
                    CTH0[idx_] = CTHs
                except:
                    raise ValueError("CBH and CTH are not the same length")
    cat["CBH"] = (("time", "height"), final_CBH)
    cat["CTH"] = (("time", "height"), final_CTH)
    cat["CBH_first"] = (("time"), CBH0)
    cat["CTH_first"] = (("time"), CTH0)
    cat["cloud_depth"] = (("time"), CTH0 - CBH0)
    cat["single_layer"] = (("time"), single_layer)
    return cat


def CTH_finder(cat: xr.Dataset, cloud_height_maximum: float = 5000):
    """
    CTH_finder finds the CTH from radar reflectivity in CLUODNET categorize files
    Input:
    cat: xarray from CLUODNET categorize file
    cloud_height_maximum: maximxum height at which CTH will be search (Default: 5000 m),
    Output:
    cat: xarray categorize input but with CTH.
    """
    top_idx, _ = utils.find_nearest_1d(cat.height.values, cloud_height_maximum)
    height = cat.height[0:top_idx].values
    CTH = np.nan * np.ones(len(cat.time))
    for i_ in np.arange(len(cat.time)):
        profile = cat.Z[i_, 0:top_idx].values
        CBH = cat.CBH[i_]
        if np.logical_and(~np.isnan(CBH), (~np.isnan(profile)).any()):
            CTH[i_] = height[~np.isnan(profile)][-1]
            if CTH[i_] <= CBH:
                CTH[i_] = np.nan

    cat["CTH"] = ("time", CTH)

    return cat


def filter_Z(cat: xr.Dataset, cbh_threshold: float = 2.5e-6, plot_flag: bool = False):
    """
    filter_Z cleans the Z matrix of low aerosol layers. It requires CLUODNET categorize files with CBH.
    Input:
    cat: xarray from CLUODNET categorize file
    cbh_threshold: attenuated beta threshold value to detect the CBH using cloudnet.CBH_attenuated()

    Output:
    cat: xarray categorize input but with cleaned Z.
    """

    if not "CBH" in cat:
        cat = CBH_attenuated(cat, threshold=cbh_threshold)

    cat["Z"] = cat.Z.where(cat.height > cat.CBH_attenuated_beta)
    if plot_flag:
        fig, axes = plt.subplots(1, 1)
        fig.set_figheight(4)
        fig.set_figwidth(15)
        cmap = mpl.cm.jet
        cat.Z.where(cat.height > cat.CBH_first).plot(
            x="time", cmap=cmap, vmin=-50, vmax=20, ax=axes
        )
        cat.CBH_first.plot(c="r", ax=axes)
    return cat


def compute_avg_in_cloud(
    Z: np.ndarray,
    v: np.ndarray,
    height: np.ndarray,
    cbh: float,
    cth: float,
    max_height: float,
) -> tuple[np.ndarray]:
    """Retrieve the Z and v mean in the given range.
    Args:
        Z (np.ndarray): reflectivity
        v (np.ndarray): vertical velocity
        height (np.ndarray): range
        cbh (float): cloud base height
        cth (float): cloud top height
        max_height (float, optional): maximum height above the CBH to perform the average. Default value None means threo.

    Returns:
        list[np.ndarray]: Z and v mean in the given range.
    """
    if max_height == -1:  # Average Over Cloud
        idc = np.logical_and(height > cbh, height < cth)
        Z_avg = np.nanmean(Z[idc])
        v_avg = np.nanmean(v[idc])
    elif max_height > 0:  # Average Over CBH and a height above CBH (and below
        # CTH)
        _, hi = utils.find_nearest_1d(height, cbh + max_height)
        idx = np.logical_and.reduce((height >= cbh, height <= hi, height <= cth))
        Z_avg = np.nanmean(Z[idx])
        v_avg = np.nanmean(v[idx])
    else:
        print("proxy_cloud must be -1 or >0")
        Z_avg = Z * np.nan
        v_avg = v * np.nan

    return Z_avg, v_avg  # FIXME


def ZdB_to_Z(Z: np.ndarray) -> np.ndarray:
    """
    Z (dBZ) to linear Z
    Z_linear = 10**(Z_log/10) x 1e-18 [mm6/m3 -> m6/m3 ]

    Args:
    - Z: units mm^6/m^3

    Returns:
    - Linear Z: units m^3
    """
    return 10 ** (Z / 10 - 18)


def filtering(
    dx_in: xr.Dataset,
    liquid_drop: bool = True,
    cbh_range: list = [450, 4000],
    lwp_range: list = [50, 150],
    lwp_rel_error_threshold: float = 50,
    similar_weather_conditions: bool = True,
    persistence: bool = True,
    force_liquid_drop_incloud: bool = True,
    preserve_time: bool = False,
    cbh_weather_conditions: bool = None,
) -> xr.Dataset:
    """_summary_

    Args:
        dx_in (xr.Dataset): categorize-classification file (merge)
        liquid_drop (bool, optional): _description_. Defaults to True.
        cbh_range (list, optional): heights where the cbh can be situated Defaults to [450, 4000].
        lwp_range (list, optional): range of values for the lwp. Defaults to [50, 150].
        lwp_rel_error_threshold (float, optional): error for lwp. Defaults to 50.
        similar_weather_conditions (bool, optional): _description_. Defaults to True.
        persistence (bool, optional): _description_. Defaults to True.
        force_liquid_drop_incloud (bool, optional): _description_. Defaults to True.
        preserve_time (bool, optional): _description_. Defaults to False.
        cbh_weather_conditions (bool, optional): _description_. Defaults to None.

    Returns:
        xr.Dataset: files that satify the conditions
    """
    # filtrado
    # - nubes liquidas
    # - rango cbh
    # - lwp>0 (e_lwp)
    # - condiciones meteo
    # - persistencia 30min
    # - si despues de filtrar, quedan menos de 10 perfiles, elimino el dia

    t = pd.to_datetime(dx_in.time.values)
    # print(t)
    h = dx_in.height.values
    # print(h)
    dx_in.model_height.values
    # print(mh)
    tc = dx_in.target_cla.values
    # print(tc)
    cbh = dx_in.cbh.values
    # print(cbh)
    cth = dx_in.cth.values
    # print(cth)
    lwp = dx_in.lwp.values
    # print(lwp)
    elwp = dx_in.lwp_error.values
    # print(elwp)
    dx_in.temperature.values
    # print(tk)
    dx_in.pressure.values
    # print(p)
    dx_in.specific_humidity.values
    # print(q)
    rr = dx_in.rainrate.values
    # print(rr)

    # TODO ADD Tk, P, Q at CLOUD BASE HEIGHT
    # tk_cbh, p_cbh, q_cbh = weather_conditions_at_cbh(cbh, mh, tk, p, q)
    # dx_in['tk_cbh'] = (['time'], tk_cbh)
    # dx_in['p_cbh'] = (['time'], p_cbh)
    # dx_in['q_cbh'] = (['time'], q_cbh)

    # index for filtering
    idx = []
    try:
        # LIQUID DROPS
        # For each profile (time), filtering is performed
        # Use TARGET CLASSIFICATION to filter out all profiles that have
        # water components other than liquid drops that may contribute to LWP:
        # drizzle, rain (2,3), ice+liquid drops (5), melting ice (6), melting
        # ice + liquid drops (7)
        # In addition, only profiles with, at least, one value for liquid
        # drop, are considered.
        idx_ld = np.ones(t.shape[0])
        if liquid_drop:  # select profiles
            for i, _t in enumerate(t):
                x = tc[:, i]
                if np.logical_and.reduce(
                    (
                        (
                            not np.logical_or.reduce(
                                (x == 2, x == 3, x == 5, x == 6, x == 7)
                            ).any()
                        )
                        and ((x == 1).any())
                    )
                ):
                    idx_ld[i] = 1
                else:
                    idx_ld[i] = 0
                del x
        # In addition, ensure no precipitation with rainrate variable
        idx_ld[rr > 0] = 0

        # MODIFY CTH

        # PERSISTENCE: if liquid_drop == TRUE
        # Se buscan perfiles consecutivos que satisfacen la condicion
        # de gota liquida durante, al menos, media hora.
        # Aflojo un poco la condicion y permito perfiles no consecutivos si
        # el lapso entre 2 no consecutivos es menor que 5 minutos.
        # obviamente, los perfiles que se intercalan entre estos dos no
        # consecutivos que, por razones obvias, no satisfacen la condición de
        # gota líquida, no se incluyen
        idx_per = idx_ld.copy()
        if persistence and liquid_drop:
            count = 1
            lp = []
            while count < t.shape[0]:
                # print("count=%i"%count)
                cp = count
                _lp = []
                if idx_ld[cp] == 1:  # si liquid drop
                    # print("comienza subset")
                    si = 1
                    last_1 = cp
                    # print("last_1=%i"%last_1)
                    while (si == 1) and (cp < t.shape[0]):  # Liquid Drop
                        if idx_ld[cp] == 1:  # si liquid drop
                            _lp.append(cp)
                            cp += 1
                            last_1 = cp
                            # print("update last_1=%i" % last_1)
                        else:  # si no liquid drop
                            if (t[cp] - t[last_1]) < np.timedelta64(300, "s"):
                                cp += 1
                            else:
                                si = 0
                    # print("termina subset")
                    lp.append(_lp)
                    count = cp + 1
                else:  # avance
                    # print("avance")
                    count += 1

            # Los subconjuntos generados deben durar, al menos, 10 minutos
            for _lp in lp:
                t_lapse = t[_lp[-1]] - t[_lp[0]]
                if t_lapse < np.timedelta64(10, "m"):
                    idx_per[_lp] = 0

        # CLOUD BASE HEIGHT RANGE
        idx_cbh = np.ones(t.shape[0])
        if isinstance(cbh_range, list):
            if len(cbh_range) == 2:
                idx = ~np.logical_and(cbh >= cbh_range[0], cbh <= cbh_range[1])
                idx_cbh[idx] = 0

        # LWP, LWP_ERROR THRESHOLDS
        idx_lwp = np.ones(t.shape[0])
        if isinstance(lwp_range, list):
            if len(lwp_range) == 2:
                idx = ~np.logical_and.reduce(
                    (
                        lwp >= lwp_range[0],
                        lwp <= lwp_range[1],
                        100 * elwp / lwp < lwp_rel_error_threshold,
                    )
                )
                idx_lwp[idx] = 0

        # ONLY LIQUID DROP INSIDE THE CLOUD
        idx_ld_incloud = np.ones(t.shape[0])
        if force_liquid_drop_incloud:
            for i, _t in enumerate(t):
                x = tc[:, i]
                ix = np.logical_and(h >= cbh[i], h <= cth[i])
                if x[ix].any():  # Existen datos incloud
                    if not (x[ix][1:] == 1).all():
                        idx_ld_incloud[i] = 0
                else:
                    idx_ld_incloud[i] = 0

        # SIMILAR METEO CONDITIONS
        # TODO usamos los perfiles que satisfacen las condiciones de gota liquida
        idx_met = np.ones(t.shape[0])
        logger.critical("Function not finished.")
        # if similar_weather_conditions:
        #     xx = idx_ld == 1
        #     idx = ~ np.logical_and(
        #         np.nanstd(tk_cbh[xx])/np.nanmean(tk_cbh[xx]) < 0.1,
        #         np.nanstd(p_cbh[xx])/np.nanmean(p_cbh[xx]) < 0.1)
        #     idx_met[idx] = 0

        # similar meteo conditions based on whole period
        # TODO: investigar si esta condicion es interesante
        """
        if cbh_weather_conditions is not None:
            # Tk, P, Q avg, sd at CBHs
            tk_cbh_avg, tk_cbh_sd = cbh_weather_conditions[0]
            p_cbh_avg, p_cbh_sd = cbh_weather_conditions[1]
            q_cbh_avg, q_cbh_sd = cbh_weather_conditions[2]

            #
            tk = dx_in.temperature.values
            p = dx_in.pressure.values
            q = dx_in.specific_humidity.values
            tk_cbh, p_cbh, q_cbh = weather_conditions_at_cbh(cbh, mh, tk, p, q)
        """

        # FILTER BY IDX
        idx = idx_ld * idx_per * idx_cbh * idx_lwp * idx_ld_incloud * idx_met
        # Hay, al menos, 10 elementos
        if len(np.argwhere(idx == 1)) >= 10:
            dx_out = dx_in.sel(time=dx_in.time[idx == 1])

            # DROP NON-NECESSARY VARIABLES
            dx_out = dx_out.drop(
                ["temperature", "pressure", "specific_humidity", "model_height"]
            )
            # REINDEX TO ORIGINAL TIME
            if preserve_time:
                dx_out = dx_out.reindex({"time": dx_in.time})
        else:
            dx_out = None
    except:
        print("SOMETHING WENT WRONG IN FILTERING")
        dx_out = None

    return dx_out



# %%
from typing import Any
from matplotlib import ticker
import numpy as np
import matplotlib.pyplot as plt

from gfatpy.utils.plot import color_list


def reflectivity(
    diameter: float | list | np.ndarray, number_per_diameter: float | list | np.ndarray
) -> tuple[float | np.ndarray, float | np.ndarray]:
    """Reflectivity factor calculation for a given diameter and number of droplets.

    Args:
        diameter (_type_): Diameter of droplets (micrometers).
        number_per_diameter (_type_): Number of droplets per diameter.

    Returns:
        _type_: _description_
    """
    if isinstance(diameter, list):
        diameter = np.array(diameter)
    if isinstance(number_per_diameter, list):
        number_per_diameter = np.array(number_per_diameter)

    linear_Z = (number_per_diameter * (diameter * 1e-3) ** 6).sum()  # en milímetros
    dBZe = 10 * np.log10(linear_Z)
    return linear_Z, dBZe


# %%
D_range = np.array([1, 10, 100, 1000])  # en micrómetros
for D in D_range:
    linear_Z, dBZe = reflectivity(D, 10)  # En milímetros
    print(f"Diameter: {D:.0e} m, Linear Z: {linear_Z:.2e}, dBZe: {dBZe:.2f} dBZ")
# %%
N = 100000
DSD = np.random.normal(10, 1, N)
diameters = np.arange(0.1, 20, 0.1)
freq, diam_bins = np.histogram(DSD, bins=diameters)
midpoints = (diam_bins[:-1] + diam_bins[1:]) / 2
linear_Z, dBZe = reflectivity(midpoints, freq)
# plot
fig, ax = plt.subplots(1, 2, figsize=(12, 6))
ax[0].plot(midpoints, linear_Z)
ax[0].set_xlabel("Droplet Diameter ($\mu m$)")
ax[0].set_ylabel("Linear Reflectivity Factor (mm^6/m^3)")
# ax[0].set_title('Linear Reflectivity Factor vs Droplet Diameter')
ax[0].grid(True)
ax[1].plot(midpoints, dBZe)
ax[1].set_xlabel("Droplet Diameter (mm)")
ax[1].set_ylabel("Reflectivity Factor (dBZ)")
# ax[1].set_title('Reflectivity Factor vs Droplet Diameter')
ax[1].grid(True)
plt.show()
# %%
N = 100000
fig, ax = plt.subplots(1, 2, figsize=(15, 6))
mu_list = [10, 20, 30, 40, 50]
colors = color_list(len(mu_list))
dBZe = np.zeros(len(mu_list))
for idx, mu in enumerate(mu_list):
    DSD = np.random.normal(mu, 1.5, N)
    diameters = np.arange(0.1, 2*mu, 0.1)
    freq, diam_bins = np.histogram(DSD, bins=diameters)
    midpoints = (diam_bins[:-1] + diam_bins[1:]) / 2
    _, dBZe[idx] = reflectivity(midpoints, freq)
    # plot
    ax[0].plot(midpoints, freq, color = colors[idx], label=f"Mean: {mu} $\mu m$")
ax[0].set_xlim(0, 60)
ax[0].set_xlabel("Droplet Diameter ($\mu m$)")
ax[0].set_ylabel("Droplet Size Distribution, $m^{-3}/\mu m$")
# ax[0].set_title('Linear Reflectivity Factor vs Droplet Diameter')
ax[0].grid(True)
ax[0].legend(loc="best", fontsize=15)

ax[1].plot(mu_list, dBZe, lw=2, marker='o', color='k')
# Color each marker according to the colors list
for idx, color in enumerate(colors):
    ax[1].plot(mu_list[idx], dBZe[idx], marker='o', markersize=20, color=colors[idx])

ax[1].set_xlabel("Droplet Diameter ($\mu m$)")
ax[1].set_ylabel("Reflectivity Factor (dBZ)")
ax[1].set_ylim(-75, -25)
# ax[1].set_title('Reflectivity Factor vs Droplet Diameter')
ax[1].grid(True)
fig.savefig("dBZe_changing_mu.png", dpi=300, bbox_inches='tight')
plt.show()
# %%
mu = 30
fig, ax = plt.subplots(1, 2, figsize=(15, 6))
N_list = 100000*np.arange(1,6)
colors = color_list(len(N_list))
dBZe = np.zeros(len(N_list))
for idx, N in enumerate(N_list):
    DSD = np.random.normal(mu, 1.5, N)
    diameters = np.arange(0.1, 2*mu, 0.1)
    freq, diam_bins = np.histogram(DSD, bins=diameters)
    midpoints = (diam_bins[:-1] + diam_bins[1:]) / 2
    _, dBZe[idx] = reflectivity(midpoints, freq)

    # plot
    ax[0].plot(midpoints, freq, color = colors[idx], label=f"{N:.1E}" +"$m^{-3}$")

ax[0].set_xlim(20, 40)
ax[0].set_xlabel("Droplet Diameter ($\mu m$)")
ax[0].set_ylabel("Droplet Size Distribution, $m^{-3}/\mu m$")
# ax[0].set_title('Linear Reflectivity Factor vs Droplet Diameter')
ax[0].grid(True)
ax[0].legend(loc="best", fontsize=12)

ax[1].plot(N_list, dBZe, lw=2, marker='o', color='k')
# Color each marker according to the colors list
for idx, color in enumerate(colors):
    ax[1].plot(N_list[idx], dBZe[idx], marker='o', markersize=20, color=colors[idx])
ax[1].set_ylim(-75, -25)
ax[1].set_xlabel("Droplet Diameter ($\mu m$)")
ax[1].set_ylabel("Reflectivity Factor (dBZ)")
# ax[1].set_title('Reflectivity Factor vs Droplet Diameter')
ax[1].grid(True)
ax[1].xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))
ax[1].ticklabel_format(style='sci', axis='x', scilimits=(0,0))

fig.savefig("dBZe_changing_N.png", dpi=300, bbox_inches='tight')
plt.show()
# %%
mu = 30
N  = 100000
fig, ax = plt.subplots(1, 2, figsize=(15, 6))
signal_list = np.arange(1,3,0.5)
colors = color_list(len(signal_list))
dBZe = np.zeros(len(signal_list))
for idx, sigma in enumerate(signal_list):
    DSD = np.random.normal(mu, sigma, N)
    diameters = np.arange(0.1, 2*mu, 0.1)
    freq, diam_bins = np.histogram(DSD, bins=diameters)
    midpoints = (diam_bins[:-1] + diam_bins[1:]) / 2
    _, dBZe[idx] = reflectivity(midpoints, freq)

    # plot
    ax[0].plot(midpoints, freq, color = colors[idx], label=f"{sigma} " +"$\mu m$")

ax[0].set_xlim(20, 40)
ax[0].set_xlabel("Droplet Diameter ($\mu m$)")
ax[0].set_ylabel("Droplet Size Distribution, $m^{-3}/\mu m$")
# ax[0].set_title('Linear Reflectivity Factor vs Droplet Diameter')
ax[0].grid(True)
ax[0].legend(loc="best", fontsize=12)

ax[1].plot(signal_list, dBZe, lw=2, marker='o', color='k')
# Color each marker according to the colors list
for idx, color in enumerate(colors):
    ax[1].plot(signal_list[idx], dBZe[idx], marker='o', markersize=20, color=colors[idx])
ax[1].set_ylim(-75, -25)
ax[1].set_xlabel("Droplet Diameter ($\mu m$)")
ax[1].set_ylabel("Reflectivity Factor (dBZ)")
# ax[1].set_title('Reflectivity Factor vs Droplet Diameter')
ax[1].grid(True)
ax[1].xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))
ax[1].ticklabel_format(style='sci', axis='x', scilimits=(0,0))

fig.savefig("dBZe_changing_sigma.png", dpi=300, bbox_inches='tight')
plt.show()

# %%
mu_list = [10, 20, 30, 40, 50]
sigma = 2
N  = 100000
# Calculate shape (k) and scale (theta) parameters
k_list = (np.array(mu_list)**2) / sigma
theta_list = (sigma / 30)*np.ones(len(k_list))

# Number of samples
size = 1000
fig, ax = plt.subplots(1, 2, figsize=(15, 6))
colors = color_list(len(k_list))
dBZe = np.zeros(len(k_list))
for idx in np.arange(len(k_list)):
    DSD = np.random.gamma(k_list[idx], theta_list[idx], N)
    diameters = np.arange(0.1, 2*mu, 0.1)
    freq, diam_bins = np.histogram(DSD, bins=diameters)
    midpoints = (diam_bins[:-1] + diam_bins[1:]) / 2
    _, dBZe[idx] = reflectivity(midpoints, freq)

    # plot
    ax[0].plot(midpoints, freq, color = colors[idx], label=f"{k_list[idx]:.2f} | {theta_list[idx]:.2f}")

# ax[0].set_xlim(20, 40)
ax[0].set_xlabel("Droplet Diameter ($\mu m$)")
ax[0].set_ylabel("Droplet Size Distribution, $m^{-3}/\mu m$")
# ax[0].set_title('Linear Reflectivity Factor vs Droplet Diameter')
ax[0].grid(True)
ax[0].legend(loc="best", fontsize=12)

ax[1].plot(k_list, dBZe, lw=2, marker='o', color='k')
# Color each marker according to the colors list
for idx, color in enumerate(colors):
    ax[1].plot(k_list[idx], dBZe[idx], marker='o', markersize=20, color=colors[idx])
# ax[1].set_ylim(-75, -25)
ax[1].set_xlabel("Droplet Diameter ($\mu m$)")
ax[1].set_ylabel("Reflectivity Factor (dBZ)")
# ax[1].set_title('Reflectivity Factor vs Droplet Diameter')
ax[1].grid(True)
ax[1].xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))
ax[1].ticklabel_format(style='sci', axis='x', scilimits=(0,0))

fig.savefig("dBZe_changing_sigma.png", dpi=300, bbox_inches='tight')
plt.show()
# %%



from pathlib import Path
import numpy as np
from pathlib import Path
from datetime import datetime
from scipy.signal import savgol_filter
import xarray as xr

from gfatpy.lidar.utils import file_manager
from gfatpy.lidar.preprocessing.lidar_preprocessing import preprocess
from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.utils.io import read_yaml_from_info


def retrieve_eta_star_profile(
    signal_T_plus45: xr.DataArray,
    signal_T_minus45: xr.DataArray,
    signal_R_plus45: xr.DataArray,
    signal_R_minus45: xr.DataArray,
    transmittance_extra_filter: float,
) -> xr.DataArray:
    """Calculate the calibration constant [eta_star] in a lidar system that is able to
    detect the T-to-R depolarization ratio.

    Args:
        signal_T_plus45 (xr.DataArray): The input vertical profile from the T channel. Calibrator angle phi=45.
        signal_T_minus45 (xr.DataArray): The input vertical profile from the R channel. Calibrator angle phi=45.
        signal_R_plus45 (xr.DataArray): The input vertical profile from the T channel. Calibrator angle phi=-45.
        signal_R_minus45 (xr.DataArray): The input vertical profile from the R channel. Calibrator angle phi=-45.
        kwargs (dict): transmittance_extra_filter [float] is the transmittance of the extra filter during depol. calibration.

    Returns:
        xr.DataArray: eta star profile.

    Notes:
        The calibration constant is calculated by the following formula:
    .. math::
        \\eta^* = \\left( \\frac{signal(T,+45)}{signal(R,+45)}\\frac{signal(T,-45)}{signal(R,-45)} \\right)^{0.5}

    References:
    Freudenthaler, V. 2016. https://doi.org/10.5194/amt-9-4181-2016
    """

    # Remove effect of the possible extra filter to avoid saturation of the s-channel.
    signal_T_plus45_with_extra_filter = signal_T_plus45
    signal_T_plus45 = signal_T_plus45_with_extra_filter / transmittance_extra_filter
    signal_T_minus45_with_extra_filter = signal_T_minus45
    signal_T_minus45 = signal_T_minus45_with_extra_filter / transmittance_extra_filter

    #   Calculate the signal ratio for the +45 position.
    delta_v45_plus = signal_R_plus45 / signal_T_plus45

    #   Calculate the signal ratio for the -45 position.
    delta_v45_minus = signal_R_minus45 / signal_T_minus45

    #   Calculate the calibration constant vertical profile.
    v_star_profile = (delta_v45_plus * delta_v45_minus) ** 0.5
    return v_star_profile


def eta_star_mean_std(
    eta_star_profile: xr.DataArray, average_range: tuple[float, float]
) -> tuple[float, float]:
    """Calculate the mean calibration constant and its standard error of the mean, from the calibration constant profile.

    Args:
    c_profile: vector
       The vertical profile of the calibration constant.
    kwargs (dict): min_range [float] is the lower vertical limit for the calculation in meters. max_range [float] is the upper vertical limit for the calculation in meters.

    Returns:
    c_mean: float
       Calibration constant's mean value (vertical axis).
    c_sem: float
       Calibration constant's standard error of the mean (vertical axis).
    """

    if "time" in eta_star_profile.dims:
        raise ValueError("eta_star_profile should have only `range` dimension.")

    #   Select the area of interest.
    eta_star_mean = (
        eta_star_profile.sel(range=slice(*average_range))
        .mean("range")
        .values.item()
    )
    eta_star_std = (
        eta_star_profile.sel(range=slice(*average_range))
        .std("range")
        .values.item()
    )

    #   Return the statistics.
    return eta_star_mean, eta_star_std


def calibration_factor_files(
    P45_fn: Path,
    N45_fn: Path,
    calib_dir: Path | str,    
    epsilon: float | None = None,
    an_calib_limits: tuple[float, float] = (750, 1250),
    pc_calib_limits: tuple[float, float] = (750, 1250),
    channel_T_transmittance: dict[str, float] | None = None,
) -> xr.Dataset:
    """It retrieves the calibration factor from two files (+45 and -45 degrees).
   
    Args:

        - P45_fn (Path): netcdf +45-degree filepath.
        - N45_fn (Path): netcdf +45-degree filepath.
        - calib_dir (Path | str): Directory to save the calibration file.
        - channel_T_transmittance dict[str, float] | None. Transmittance of the extra filter for each channel. Defaults means values taken from lidar info.
        - epsilon (float | None, optional): Misalignment angle of the calibrator. Defaults to None means zero.
        - an_calib_limits (tuple[float, float], optional): Mininum amd maximum range to performe the analaog calibration average. Defaults to (1500, 3000).
        - pc_calib_limits (tuple[float, float], optional): Mininum amd maximum range to performe the photoncounting(gluing) calibration average. Defaults to (2500, 4000).

    Raises:
        - FileNotFoundError: +45-degrees file not found.
        - FileNotFoundError: -45-degrees file not found.
        - FileNotFoundError: Calibration directory not found.
        - ValueError: detection mode not recognized.

    Returns:
        - xr.Dataset: calibration factor dataset.
    """    
    if not P45_fn.is_file():
        raise FileNotFoundError(f"{P45_fn} not found.")

    if not N45_fn.is_file():
        raise FileNotFoundError(f"{N45_fn} not found.")

    if isinstance(calib_dir, str):
        calib_dir = Path(calib_dir)

    if not calib_dir.is_dir():
        raise FileNotFoundError(f"{calib_dir} not found.")

    #Get correct lidar info using the filename    
    lidar_nick, _, _, _, _, calib_date  = file_manager.filename2info(P45_fn.name)

    global INFO 
    INFO = read_yaml_from_info(lidar_nick, calib_date)

    lidar_name: str = LIDAR_INFO["metadata"]["nick2name"][lidar_nick]

    P45 = preprocess(P45_fn)
    N45 = preprocess(N45_fn)

    calib_dict = {}
    for telescope_ in INFO[
        "polarized_channels"
    ].keys():

        calib_dict[telescope_] = {}

        for wavelength_ in INFO["polarized_channels"][
            telescope_
        ].keys():

            calib_dict[telescope_][wavelength_] = {}

            for mode_ in INFO["polarized_channels"][
                telescope_
            ][wavelength_].keys():
                if mode_ == 'a':
                    average_range = an_calib_limits
                else:
                    average_range = pc_calib_limits

                # Define the channel names
                channel_T = INFO["polarized_channels"][
                    telescope_
                ][wavelength_][mode_]["T"]

                channel_R = INFO["polarized_channels"][
                    telescope_
                ][wavelength_][mode_]["R"]

                # Check if the channel is available in both P45 and N45.
                if (
                    channel_T not in P45.channel.values
                    or channel_R not in P45.channel.values
                ):
                    continue

                if (
                    channel_T not in N45.channel.values
                    or channel_R not in N45.channel.values
                ):
                    continue

                calib_dict[telescope_][wavelength_][mode_] = {}

                # Apply smoothing filter
                signal_T_P45 = xr.apply_ufunc(
                    savgol_filter,
                    P45[f"signal_{channel_T}"].mean("time"),
                    11,
                    3,
                    dask="allowed",
                )
                signal_R_P45 = xr.apply_ufunc(
                    savgol_filter,
                    P45[f"signal_{channel_R}"].mean("time"),
                    11,
                    3,
                    dask="allowed",
                )
                signal_R_N45 = xr.apply_ufunc(
                    savgol_filter,
                    N45[f"signal_{channel_R}"].mean("time"),
                    11,
                    3,
                    dask="allowed",
                )
                signal_T_N45 = xr.apply_ufunc(
                    savgol_filter,
                    N45[f"signal_{channel_T}"].mean("time"),
                    11,
                    3,
                    dask="allowed",
                )
                
                if channel_T_transmittance is None:
                    transmittance_extra_filter = INFO["depolarization_calibration_transmittance"][channel_T[:-1]]
                else:
                    raise ValueError(f"channel_T_transmittance of channel {channel_T[:-1]} not found in lidar info config file. Dictionary: 'depolarization_calibration_transmittance'.")

                # Calculate the calibration constant vertical profile.
                eta_star_profile = retrieve_eta_star_profile(
                    signal_T_P45,
                    signal_T_N45,
                    signal_R_P45,
                    signal_R_N45,
                    transmittance_extra_filter=transmittance_extra_filter,
                )
                calib_dict[telescope_][wavelength_][mode_]["profile"] = eta_star_profile
                eta_star_mean, eta_star_std = eta_star_mean_std(
                    eta_star_profile, average_range=average_range
                )

                # Calibrator rotation, epsilon [Freudenthaler, V. (2016)., Eq. 194, 195]
                # average over calibration height interval
                gain_ratio_p45 = signal_T_P45 / signal_R_P45
                gain_ratio_n45 = signal_T_N45 / signal_R_N45

                ranges = P45.range.values

                match mode_:
                    case "a":
                        idx_avg_ranges = (ranges >= an_calib_limits[0]) & (
                            ranges <= an_calib_limits[1]
                        )
                    case "p":
                        idx_avg_ranges = (ranges >= pc_calib_limits[0]) & (
                            ranges <= pc_calib_limits[1]
                        )
                    # case "g":
                    #     ...
                    # TODO: Gluing gna be implemented in the future by receiving extra argument
                    case _:
                        raise ValueError(f"Mode {mode_} not recognized")

                Y = (gain_ratio_p45 - gain_ratio_n45) / (
                    gain_ratio_p45 + gain_ratio_n45
                )
                Y_avg = np.nanmean(Y[idx_avg_ranges])
                Y_std = np.nanstd(Y[idx_avg_ranges])
                if epsilon is None:
                    epsilon = (
                        (180 / np.pi) * 0.5 * np.arcsin(np.tan(0.5 * np.arcsin(Y_avg)))
                    )

                epsilon_err = (
                    (180 / np.pi)
                    * 0.5
                    * abs(
                        0.5 * np.arcsin(np.tan(0.5 * np.arcsin(Y_avg + Y_std)))
                        - 0.5 * np.arcsin(np.tan(0.5 * np.arcsin(Y_avg - Y_std)))
                    )
                )

                calib_dict[telescope_][wavelength_][mode_]["values"] = [
                    eta_star_mean,
                    eta_star_std,
                    Y_avg,
                    Y_std,
                    epsilon,
                    epsilon_err,
                ]

                calib_dict[telescope_][wavelength_][mode_]["signals"] = [
                    signal_T_N45,
                    signal_T_P45,
                    signal_R_P45,
                    signal_R_N45,
                    gain_ratio_p45,
                    gain_ratio_n45,
                ]

    # Create dataset
    calib_dataset = xr.Dataset()

    channels = []
    for telescope_ in calib_dict.keys():
        for wavelength_ in calib_dict[telescope_].keys():
            for mode_ in calib_dict[telescope_][wavelength_].keys():
                # Add the channel name to a list
                channels.append(f"{wavelength_}{telescope_[0]}{mode_}")

                # Define the key for the profile
                key_profile = f"eta_star_profile_{wavelength_}{telescope_[0]}{mode_}"
                calib_dataset[key_profile] = calib_dict[telescope_][wavelength_][mode_][
                    "profile"
                ]

                # Get the depolarization values
                (
                    eta_star_mean,
                    eta_star_std,
                    Y_avg,
                    Y_std,
                    epsilon,
                    epsilon_err,
                ) = calib_dict[telescope_][wavelength_][mode_]["values"]

                # Define xarray dataarrays
                calib_dataset[
                    f"eta_star_mean_{wavelength_}{telescope_[0]}{mode_}"
                ] = xr.DataArray(
                    eta_star_mean,
                    dims=[],
                    attrs={
                        "long_name": f"range-average of {key_profile}",
                        "min_range_m": average_range[0],
                        "max_range_m": average_range[1],
                    },
                )
                calib_dataset[
                    f"eta_star_standard_deviation_{wavelength_}{telescope_[0]}{mode_}"
                ] = xr.DataArray(
                    eta_star_std,
                    dims=[],
                    attrs={
                        "long_name": f"range standard-deviation of {key_profile}",
                        "min_range_m": average_range[0],
                        "max_range_m": average_range[1] 
                    },
                )

                calib_dataset[
                    f"Y_average_{wavelength_}{telescope_[0]}{mode_}"
                ] = xr.DataArray(
                    Y_avg,
                    dims=[],
                    attrs={
                        "long_name": f"range standard-deviation of {key_profile}",
                        "min_range_m": average_range[0],
                        "max_range_m": average_range[1]
                    },
                )
                calib_dataset[
                    f"Y_standard_deviation_{wavelength_}{telescope_[0]}{mode_}"
                ] = xr.DataArray(
                    Y_std,
                    dims=[],
                    attrs={
                        "long_name": f"range standard-deviation of {key_profile}",
                        "min_range_m": average_range[0],
                        "max_range_m": average_range[1]
                    },
                )

                calib_dataset[
                    f"epsilon_{wavelength_}{telescope_[0]}{mode_}"
                ] = xr.DataArray(
                    epsilon,
                    dims=[],
                    attrs={
                        "long_name": f"range standard-deviation of {key_profile}",
                        "min_range_m": average_range[0],
                        "max_range_m": average_range[1]
                    },
                )

                calib_dataset[
                    f"epsilon_error_{wavelength_}{telescope_[0]}{mode_}"
                ] = xr.DataArray(
                    epsilon_err,
                    dims=[],
                    attrs={
                        "long_name": f"range standard-deviation of {key_profile}",
                        "min_range_m": average_range[0],
                        "max_range_m": average_range[1]
                    },
                )

                signals_order = [
                    "signal_T_N45",
                    "signal_T_P45",
                    "signal_R_P45",
                    "signal_R_N45",
                    "gain_ratio_p45",
                    "gain_ratio_n45",
                ]

                for signal, signal_name in zip(calib_dict[telescope_][wavelength_][mode_]["signals"], signals_order):  # type: ignore
                    calib_dataset[
                        f"{signal_name}_{wavelength_}{telescope_[0]}{mode_}"
                    ] = signal

    # Channels is not a coordinate yet, just a utility to read the file
    calib_dataset["channels"] = channels

    # Add time
    calib_datetime_str = datetime.strftime(calib_date, "%Y%m%d_%H%M")
    calib_dataset.attrs["calibration_datetime"] = calib_datetime_str
    calib_dataset.attrs["system"] = lidar_name

    calib_filename = f"{lidar_nick}_eta-star_{calib_datetime_str}.nc"
    calib_dataset.to_netcdf(calib_dir / calib_filename)
    return calib_dataset



def eta_star_reader(filepath: Path) -> xr.Dataset:
    """It loads the eta_star data from files generated with `gfatpy.lidar.depolarization.calibration.calibration_factor_files`.

    Args:

        - filepath (Path): Filepath output from `gfatpy.lidar.depolarization.calibration.calibration_factor_files`.

    Raises:

        - FileNotFoundError: File not found.

    Returns:

        - xr.Dataset: eta_star in dataset. 
    """    
    if not filepath.is_file():
        raise FileNotFoundError(f"{filepath} not found.")

    eta_star_dataset = xr.open_dataset(filepath)
    return eta_star_dataset


import datetime as dt
from datetime import datetime
from pathlib import Path
from typing import Any
import numpy as np
import xarray as xr
import pandas as pd
from scipy.stats import pearsonr
import matplotlib.pyplot as plt


from gfatpy.lidar.quality_assurance.dead_time import binning, get_valid_an_pc_values, cost_function


def tau_cmax_plot(c_max_range: list[float] | np.ndarray[Any, np.dtype[np.float64]],
                  optimal_taus: list[float] | np.ndarray[Any, np.dtype[np.float64]],
                  pc_signal_channel: str,
                  target_date: str | dt.date,
                  savefig: bool = True,
                  plot_dir: Path = Path.cwd()
                  ):
    
    """Plots the results of optimal_taus vs C_max.
    
    Args:
        c_max_range (np.array): pc threshold candidates.
        optimal_taus (np.array): Optimal tau values.
        pc_signal_channel (str): Channel name.
        target_date (str): Target date.
        savefig (bool): Save the plot. Default=False.
        plot_dir (Path): Path to save the plot. Default=Path.cwd().
        
    Returns:
        None
    """
    if isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y%m%d")

    fig,ax = plt.subplots(figsize=(6,4))
    ax.plot(c_max_range,optimal_taus,  marker='.',linestyle='-',color='b',label=r'Optimal $\tau$')

    ax.set_ylabel(r'Optimal $\tau$ [ns]',fontsize=10)
    ax.set_xlabel(r'$C_{max}$ [MHz]',fontsize=10)
    ax.set_title('Optimal tau vs C_max',fontsize=10)
    ax.legend(fontsize=12)
    xtick_locs, xtick_labels = ax.get_xticks(), ax.get_xticklabels()
    ytick_locs, ytick_labels = ax.get_yticks(), ax.get_yticklabels()
    ax.set_xticks(xtick_locs)
    ax.set_xticklabels(xtick_labels,fontsize=10)
    ax.set_yticks(ytick_locs)
    ax.set_yticklabels(ytick_labels,fontsize=10)
    if savefig==True:
        target_date_str = target_date.strftime("%y%m%dT%H%M%S")
        fig.savefig(plot_dir / f'tau_cmax_{pc_signal_channel}_{target_date_str}.png')


def dead_time_corr_study_prf(signal_an: xr.DataArray, 
                             signal_pc: xr.DataArray,
                             taus_cmax: pd.DataFrame,
                             pc_signal_channel: str, 
                             c_max_treshold: float = 100,
                             max_corr_range: float = 350,
                             plot_crop_range: tuple = (0,2000),
                             reference_range: float = 500,
                             debugging: bool = False, 
                             save_debug_plot: bool = False, 
                             plot_dir: Path | str | None = None):
    
    """
    Study the effect of the dead time correction in the correlation range and 
    select the best tau and C_max pair.
    
    Args:
        signal_an (xr.Dataset): Analog signal.
        signal_pc (xr.Dataset): Photon counting signal.
        taus_cmax (pd.DataFrame): Dead time values and pc thresholds candidates.
        pc_signal_channel (str): Photon counting channel name.
        c_max_treshold (float): Photon counting threshold in MHz for range in which to calculate corr. Default=100.
        max_corr_range (float): Maximum correlation range. Default=350.
        plot_crop_range (tuple): Plot crop range to limit plot and calculation range. Default=(0,2000).
        reference_range (float): Reference range in whicih to normalise an signal to pc. Default=500.
        debugging (bool): Debugging mode with additonal plots. Default=False.
        save_debug_plot (bool): Save the plot. Default=False.
        plot_dir (str | None): Path to save the plot. Default=None means current working directory.
        
    Returns:
        best_tau (float): Optimal dead time value.
        best_cmax (float): Optimal pc threshold value.
    """
    
    """Crop the signals using plot_crop_range"""

    if isinstance(plot_dir, str):
        plot_dir = Path(plot_dir)
    if plot_dir is None:
        plot_dir = Path.cwd()

    # Crop pc and an signals for a better plot in the near-medium range and select a time range (nightime): 
    #plot_crop_range = (0,2000)    # dt correction effect important when the signal i sstrong -> near range
    raw_pc_prf = signal_pc.sel(range=slice(*plot_crop_range))
    raw_an_prf = signal_an.sel(range=slice(*plot_crop_range))
    

    """Apply dt correction by hand using taus and C_max from dead_time_finder_channel_victor"""
    # 1) Load the (taus,C_max) pairs and correct of dt:
    C_max = taus_cmax['c_max'].values                                          
    taus = taus_cmax[f'dead_time_{pc_signal_channel}'].values
    # 2) Dead time correction by hand:
    dt_pc_signals = []
    for t in taus:
        dt_pc = raw_pc_prf / (1 - raw_pc_prf * (t*1e-3))   # dt correction with tau in us
        dt_pc_signals.append(dt_pc)
    """Apply C_max treshold and select the valid zone after overlap peak"""
    # 1) To select zone after the overlap, we select range after the max (overlap)
    max_idx = np.where(raw_pc_prf.values==np.max(raw_pc_prf.values))[0][0] # type: ignore
    min_range = raw_pc_prf.range.values[max_idx]
    max_range = plot_crop_range[1]
    cut_raw_pc_prf = raw_pc_prf.sel(range=slice(min_range,max_range))
    # 2) Now we apply C_max treshold:
    c_max = c_max_treshold                     # MHz
    mask = (cut_raw_pc_prf<c_max).values       # boolean mask
    cut_raw_pc_prf = cut_raw_pc_prf[mask]      # masked array that fulfill c<c_max and after the overlap

    if debugging==True:
        # Plot 1 -> effect of the dt correction in the raw pc signals
        fig_1,ax_1=plt.subplots(figsize=(9,5))        
        ax_1.plot(raw_an_prf.range,raw_an_prf, color ='black', label='raw_an') #type: ignore
        ax_1.plot(raw_pc_prf.range,raw_pc_prf, color ='blue', label='raw_pc') #type: ignore
        #ax_1.plot(n_raw_an.range,n_raw_an, color='black', label='norm_an')
        #ax_1.plot(dt_pc_signals[0].range,dt_pc_signals[0], label=f'tau = {taus[0]}')  # equivalent to raw signal
        #ax_1.plot(dt_pc_signals[1].range,dt_pc_signals[1], label=f'tau = {taus[1]}')  # really bad tau value -> bad result
        ax_1.plot(dt_pc_signals[2].range,dt_pc_signals[2], label=f'tau = {taus[2]}')
        ax_1.plot(dt_pc_signals[3].range,dt_pc_signals[3], label=f'tau = {taus[3]}')
        ax_1.plot(dt_pc_signals[4].range,dt_pc_signals[4], label=f'tau = {taus[4]}')
        ax_1.plot(dt_pc_signals[5].range,dt_pc_signals[5], label=f'tau = {taus[5]}')
        ax_1.plot(dt_pc_signals[6].range,dt_pc_signals[6], label=f'tau = {taus[6]}')
        ax_1.hlines(y=c_max,xmin=0,xmax=plot_crop_range[1], color='red', linestyle='--', label=f'Treshold: {c_max} [MHz]')
        ax_1.fill_between(cut_raw_pc_prf.range,y1=cut_raw_pc_prf,y2=c_max, alpha=0.3, color='green', label='Valid corr. zone')
        ax_1.set_xlim(0,plot_crop_range[1])
        ax_1.legend(fontsize=9)

        if save_debug_plot==True:
            fig_1.savefig(plot_dir / f'dt_effect_and_valid_corr_range_{pc_signal_channel}_{raw_pc_prf.time.dt.strftime("%y%m%d").values}.png')

    """Corr Calculations: using scipy.signal.pearsonr and a LS fit to obtain slope and offset"""
    # 1) Crop the signals to the correlation range:
    min_corr_range = cut_raw_pc_prf.range.values[0]  # first range of the already cut signal
    max_corr_range = max_corr_range                  # max range for corr. as a user parameter
    if min_corr_range>max_corr_range:
        min_corr_range, max_corr_range = 150,350
    raw_an_corr_range = raw_an_prf.sel(range=slice(min_corr_range,max_corr_range))  # not normalised
    corr_cut_pc_signals = [s.sel(range=slice(min_corr_range,max_corr_range)) for s in dt_pc_signals]
    # 2) Correlation calculations:
    r_pearson = []
    p_pearson = []
    slopes = []
    offset = []
    for i,s in enumerate(corr_cut_pc_signals):
        # 2.1) Correlation with scipy.signal.pearsonr:  (Not really necessary)
        # Use the raw an signal for corr. calculations
        #print(s.values.shape,raw_an_corr_range.values.shape)
        r,p = pearsonr(s.values,raw_an_corr_range.values)
        r_pearson.append(r)
        p_pearson.append(p)
        # 2.2) Correlation with LS fit:
        fit_params = np.polyfit(raw_an_corr_range.values,s.values,1)
        m,o = fit_params
        slopes.append(m)
        offset.append(o)
        if debugging==True:
        # Plot 2 -> correlation figures:
            fig_2,ax_2 = plt.subplots(figsize=(4,4))
            ax_2.plot(raw_an_corr_range.values,s.values,'b.',label = f'tau = {taus[i]} [ns]' )
            ax_2.plot(raw_an_corr_range.values,np.polyval(fit_params,raw_an_corr_range.values), label = f'y={m:.3f}x + {o:.3f}')
            ax_2.legend()
            if save_debug_plot==True:
                fig_2.savefig(plot_dir / f'corr_{pc_signal_channel}_{taus[i]}_{C_max[i]}_{raw_pc_prf.time.dt.strftime("%y%m%d").values}.png')
    
    # 3) Select the best tau and C_max pair that gave closest to zero offset.
    best_idx = np.argmin(np.abs(offset))   # Which correlation gave closest to zero offset
    best_tau, best_cmax = taus[best_idx], C_max[best_idx]

    """Create df with results for all c_max and taus"""
    result_df = pd.DataFrame()
    result_df['c_max'] = C_max
    result_df['tau'] = taus
    result_df['r_pearson'] = np.array(r_pearson).round(6)
    result_df['p_pearson'] = p_pearson
    result_df['slopes'] = slopes
    result_df['offsets'] = offset
    #result_df['r_spearman'] = r_spearman
    #result_df['p_spearman'] = p_spearman
    #result_df = result_df.round(6)
    if debugging==True:
        print(result_df)
        if save_debug_plot==True:
            result_df.to_csv(plot_dir / f'dt_study_results_{pc_signal_channel}_{raw_pc_prf.time.dt.strftime("%y%m%d").values}.csv')
    
    """Final plot in the correlation range with the best tau and C_max and the dt_effects"""
    if debugging==True:
        # Plot 3 -> dt effect in the corr range and the best tau:
        # Normalise an to pc for better plot
        # Divide by the own signal in that value to transport it to 0,1 and then scale with the pc signal.
        # NOTE: if the selected range doesn't exactly exits, we select the 'nearest'.
        n_raw_an = (raw_an_prf / raw_an_prf.sel(range=reference_range,method='nearest'))* raw_pc_prf.sel(range=reference_range,method='nearest')
        n_raw_an_corr_range = n_raw_an.sel(range=slice(min_corr_range,max_corr_range))
        
        fig_3,ax_3= plt.subplots(figsize=(8,5))
        
        for i,s in enumerate(corr_cut_pc_signals[2:]):
            ax_3.plot(s.range,s,label=f'corr = {r_pearson[i+2].round(6)} ; tau = {taus[i+2]} [ns]')
            if i==best_idx:
                ax_3.fill_between(s.range,y1=s-2,y2=s+2,alpha=0.3,color='lime',label=f'Best tau = {taus[i+2]} [ns] and C_max = {C_max[i+2]} [MHz]')

        ax_3.plot(corr_cut_pc_signals[0].range,corr_cut_pc_signals[0],color='blue',label='raw_pc')
        ax_3.plot(n_raw_an_corr_range.range,n_raw_an_corr_range,color='black',label='norm_an')
        ax_3.set_ylabel('Raw pc signal [MHz]',fontsize=10)
        ax_3.set_xlabel('Range',fontsize=10)
        xtick_locs, xtick_labels = ax_3.get_xticks(), ax_3.get_xticklabels()
        ytick_locs, ytick_labels = ax_3.get_yticks(), ax_3.get_yticklabels()
        ax_3.set_xticks(xtick_locs)
        ax_3.set_xticklabels(xtick_labels,fontsize=10)
        ax_3.set_yticks(ytick_locs)
        ax_3.set_yticklabels(ytick_labels,fontsize=10)
        ax_3.legend(fontsize=9)

        if save_debug_plot==True:
            fig_3.savefig(plot_dir / f'dt_study_results_{pc_signal_channel}_{raw_pc_prf.time.dt.strftime("%y%m%d").values}.png')

    return best_tau, best_cmax


def dead_time_assesment_by_channel(
        lidar: xr.Dataset,
        an_signal_channel: str,
        pc_signal_channel: str,
        ini_corr_time: str | dt.datetime, 
        end_corr_time: str | dt.datetime,
        savefig: bool,
        an_threshold: float = 0.01,
        debugging: bool = True,
        tau_range: list[float] | np.ndarray[Any, np.dtype[np.float64]] = np.arange(2., 10., 0.01),
        c_max_range: list[float] | np.ndarray[Any, np.dtype[np.float64]] = np.arange(10., 100., 10.),
        pc_bin_width: float = 1.0,
        output_dir: Path = Path.cwd()
        ):
    
    """Find the optimal dead time value for a given channel by applying different 
    pc thresholds to obtain dt and then studying corraltion between an and dt corrected pc signals.

    Args:
        lidar (Path): Lidar data file.
        an_signal_channel (str): Analog channel name.
        pc_signal_channel (str): Photon counting channel name.
        ini_corr_time (str | dt.datetime): initial time of the range from which the profile for corr study is selected
        end_corr_time (str | dt.datetime): final time of the range from which the profile for corr study is selected
        savefig (bool): Save the plot. Default=False.
        an_threshold (float): Analog threshold in mV. Default=0.01.
        debugging (bool): Debugging mode. Default=True. Wheter to show additional correlation plots or not.
        tau_range (list | np.darray): Candidate dead time values. Default=np.arange(2, 10, 0.01).
        c_max_range (list | np.darray): Candidate pc thresholds. Default=np.arange(0, 100, 10).
        pc_bin_width (float): Width of the bin. Default=1.0. In MHz.
        output_dir (Path): Path to save the results and plots. Default=Path.cwd().
        

    Returns:
        best_tau (float): Optimal dead time value.
        best_cmax (float): Optimal pc threshold value.
    """

    if isinstance(tau_range, list):
        tau_range = np.array(tau_range)

    target_date = lidar.time.values[0].astype(str).split("T")[0].replace("-", "")
    signal_an = lidar[f"signal_{an_signal_channel}"]
    signal_pc = lidar[f"signal_{pc_signal_channel}"]


    optimal_taus = []
    J_values = []
    for c in c_max_range:
        # 1) Treshold condition:
        valid_signal_an,valid_signal_pc = get_valid_an_pc_values(signal_an=signal_an,
                                                            signal_pc=signal_pc,
                                                            pc_threshold=c,
                                                            an_threshold=an_threshold)
        #print('valid values ok')

        # 2) Binning:
        median_pc, median_an, std_an = binning(signal_an = valid_signal_an, 
                                             signal_pc = valid_signal_pc, 
                                             pc_signal_channel= pc_signal_channel,
                                             target_date=target_date,
                                             pc_binning_range=(0,c),
                                             pc_bin_width=pc_bin_width, # pc_bin_width in MHz
                                             plot_dir = output_dir,
                                             savefig = savefig )


        #print('binning ok')
        # 3) Minimising and obtaining the optimal tau:
        J, optimal_tau = cost_function(tau_range = tau_range, 
                                       median_an = median_an, 
                                       median_pc = median_pc, 
                                       std_an = std_an, 
                                       pc_signal_channel = pc_signal_channel,
                                       target_date = target_date, 
                                       savefig = savefig,
                                       plot_dir = output_dir)
        J_values.append(J)
        optimal_taus.append(optimal_tau)
        #print('minimisation ok')

    # 4) Plotting:
    tau_cmax_plot(
        c_max_range= c_max_range,
        optimal_taus= optimal_taus,
        pc_signal_channel= pc_signal_channel,
        target_date= target_date,
        savefig= savefig,
        plot_dir= output_dir)

    taus_cmax = pd.DataFrame()
    taus_cmax['c_max'] = c_max_range
    taus_cmax[f'dead_time_{pc_signal_channel}'] = optimal_taus
    
    """Study which pair (tau,C_max) is the best for the dead time correction using the 1st profile"""
    if isinstance(ini_corr_time, dt.datetime):
        ini_corr_time = ini_corr_time.strftime("%Y-%m-%dT%H:%M:%S")
    if isinstance(end_corr_time, dt.datetime):
        end_corr_time = end_corr_time.strftime("%Y-%m-%dT%H:%M:%S")

    raw_pc = signal_pc.sel(time=slice(ini_corr_time,end_corr_time))
    raw_an = signal_an.sel(time=slice(ini_corr_time,end_corr_time))
    raw_pc_prf = raw_pc[0]     # 1st signal in the time range -> individual profile
    raw_an_prf = raw_an[0]     # 1st signal in the time range -> individual profile
    best_tau, best_cmax = dead_time_corr_study_prf(signal_an= raw_an_prf, 
                             signal_pc = raw_pc_prf,
                             taus_cmax = taus_cmax,
                             pc_signal_channel = pc_signal_channel, 
                             c_max_treshold = 100,
                             max_corr_range =  350,
                             plot_crop_range = (0,2000),
                             reference_range = 500,
                             debugging = debugging)

    return best_tau, best_cmax
    
    """Now we could use this for several days and chek that the best result is always for 50 MHz"""
    """Then average the tau of the different days similar to the 2009 article"""





#!/usr/bin/env python
# -*- coding: utf8 -*-

"""
various utilities
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import re
import os
import sys
from typing import Callable

import scipy as sp
import numpy as np
import pandas as pd
import xarray as xr
import datetime as dt
from scipy import stats
from loguru import logger
import statsmodels.api as sm
from statsmodels.stats.stattools import durbin_watson
from scipy.stats import anderson


"""
"""


""" NUMERICAL
"""


def normalize(x):  # , exclude_inf=True):
    """Normalize data in a 1-D array: [0, 1]

    Args:
        x ([type]): [description]
        exclude_inf (bool, optional): [description]. Defaults to True.

    Returns:
        [type]: [description]
    """

    n = np.zeros(len(x)) * np.nan
    try:
        idx_fin = np.logical_and(x != -np.inf, x != np.inf)
        x0 = x[idx_fin]
        n[idx_fin] = (x0 - np.nanmin(x0)) / (np.nanmax(x0) - np.nanmin(x0))
    except Exception as e:
        print("ERROR in normalize. %s" % str(e))

    return n


def interp_nan(y, last_value=0):
    """

    Parameters
    ----------
    y: array
        1d array with nans

    Returns
    -------
    y: array
        interpolated array

    """
    # interpolation fails if values at extreme of array are nan
    if np.isnan(y[-1]):
        y[-1] = last_value
    if np.isnan(y[0]):
        y[0] = last_value

    # bool of nans
    nans = np.isnan(y)
    # funcion lambda que genera un vector de indices a partir de un array logico
    x = lambda z: z.nonzero()[0]
    # def interpolate function
    f = sp.interpolate.interp1d(x(~nans), y[~nans], kind="cubic")
    # do interpolation
    y[nans] = f(x(nans))

    return y


def linear_regression(x, y):
    """
    y = a*x + b

    :param x: abscisa. 1-D array.
    :param y: ordenada. 1-D array.
    :return:
    """
    try:
        x = np.asarray(x)
        y = np.asarray(y)

        # Clean data
        idx = np.logical_and(~np.isnan(x), ~np.isnan(y))
        x_train = x[idx]
        y_train = y[idx]

        # Regression
        lr = stats.linregress(x_train, y_train)
        slope = float(lr.slope)  # type: ignore
        intercept = float(lr.intercept)  # type: ignore
        rvalue = float(lr.rvalue)  # type: ignore
    except Exception as e:
        print("ERROR. In linear_regression. %s" % str(e))
        slope = np.nan
        intercept = np.nan
        rvalue = np.nan

    return slope, intercept, rvalue


def linrest(x, y):
    n = len(y)
    dofreedom = n - 2
    z, _ = np.polyfit(x, y, 1, cov=True)
    p = np.poly1d(z)
    yp = p(x)  # predicted y values based on fit
    slope = z[0]
    intercept = z[1]
    r2 = np.corrcoef(x, y)[0][1] ** 2
    # regression_ss = np.sum( (yp-np.mean(y))**2)
    residual_ss = np.sum((y - yp) ** 2)
    slope_pm = np.sqrt(residual_ss / (dofreedom * np.sum((x - np.mean(x)) ** 2)))
    intercept_pm = slope_pm * np.sqrt(np.sum(x**2) / n)
    # s = np.sqrt(residual_ss/dofreedom)
    # F = regression_ss/s**2

    return slope, slope_pm, intercept, intercept_pm, r2


def residuals(meas, pred):
    """
    Residuals: J = (1/n)*sum{ [(meas-pred)/std(meas)]**2 }

    :param meas: 1-D array. measurement
    :param pred: 1-D array. prediction
    :return:
    """
    n = len(meas)
    try:
        sigma = np.nanstd(meas)
        J = np.nansum(((meas - pred) / sigma) ** 2) / n
    except Exception as e:
        print("ERROR. In cost_function %s" % str(e))
        J = np.nan
    return J


"""
ARRAYS
"""


def unique(array):
    """
    Get unique and non-nan values of a 1D array
    :param array:
    :return: array of unique values
    """
    try:
        unq = np.unique(array[~np.isnan(array)])
    except Exception as e:
        unq = np.nan
        print("Error: getting unique values of an array. %s" % str(e))

    return unq


def find_nearest_1d(array, value):  # TODO: Isn't this the same as np.searchsorted?
    """
    Find nearest value in a 1-D array
    :param array:
    :param value:
    :return:
    """
    array = np.asarray(array)
    if np.logical_and(~np.isnan(value), ~np.isnan(array).all()):
        idx = (np.abs(array - value)).argmin()
        nearest = array[idx]
    else:
        idx = np.nan
        nearest = np.nan

    return idx, nearest


"""
OTHERS
"""


def check_dir(dir_name):
    """
    Check if a directory exists and is writable
    """

    return os.access(dir_name, os.W_OK)


def welcome(prog_name, prog_version):
    """print informations about the code"""

    print("starting {} v{}".format(prog_name, prog_version))
    print()


def print_progress(iteration, total, prefix="", suffix="", decimals=2, bar_length=100):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : number of decimals in percent complete (Int)
        bar_length  - Optional  : character length of bar (Int)
    """

    filled_length = int(round(bar_length * iteration / float(total)))
    percents = round(100.00 * (iteration / float(total)), decimals)
    progress_bar = "*" * filled_length + "-" * (bar_length - filled_length)
    sys.stdout.write(
        "\r%s |%s| %s%s %s" % (prefix, progress_bar, percents, "%", suffix)
    )
    sys.stdout.flush()
    if iteration == total:
        sys.stdout.write("\n")
        sys.stdout.flush()


def numpy_to_datetime(numpy_date):
    """
    Converts a numpy datetime64 object to a python datetime object
    Input:
      date - a np.datetime64 object
    Output:
      DATE - a python datetime object
    """
    # timestamp = ((date - np.datetime64('1970-01-01T00:00:00'))
    #             / np.timedelta64(1, 's'))

    if isinstance(numpy_date, np.ndarray):
        numpy_date = numpy_date[0]
    else:
        numpy_date = numpy_date

    try:
        timestamp = dt.datetime.utcfromtimestamp(numpy_date.tolist() / 1e9)
    except Exception as e:
        timestamp = None
        print(str(e))

    return timestamp


def datetime_np2dt(numpy_date):
    """
    Converts a numpy datetime64 object to a python datetime object
    Input:
      date - a np.datetime64 object
    Output:
      DATE - a python datetime object
    """
    # timestamp = ((date - np.datetime64('1970-01-01T00:00:00'))
    #             / np.timedelta64(1, 's'))

    try:
        timestamp = pd.Timestamp(numpy_date).to_pydatetime()
    except Exception as e:
        timestamp = None
        logger.error(str(e))
        raise e
    return timestamp


def str_to_datetime(date_str: str) -> dt.datetime:
    """

    Parameters
    ----------
    date_str: str
        date in string format (see possible formats below)

    Returns
    -------

    """
    assert isinstance(date_str, str), "date_str must be String Type"

    formats = [
        (r"\d{4}\d{2}\d{2}T\d{2}\d{2}\d{2}", "%Y%m%dT%H%M%S"),
        (r"\d{4}\d{2}\d{2}_\d{2}\d{2}\d{2}", "%Y%m%d_%H%M%S"),
        (r"\d{4}\d{2}\d{2}T\d{2}\d{2}", "%Y%m%dT%H%M"),
        (r"\d{4}\d{2}\d{2}_\d{2}\d{2}", "%Y%m%d_%H%M"),
        (r"\d{4}\d{2}\d{2}T\d{2}", "%Y%m%dT%H"),
        (r"\d{4}\d{2}\d{2}_\d{2}", "%Y%m%d_%H"),
        (r"\d{4}\d{2}\d{2}", "%Y%m%d"),
        (r"\d{4}\d{2}", "%Y%m"),
        (r"\d{4}", "%Y"),
    ]

    i = 0
    match = False
    date_format = ""
    date_dt = None
    while not match:
        if i < len(formats):
            candidate = re.search(formats[i][0], date_str)
            if candidate is not None:
                date_format = formats[i][1]
                match = True
            else:
                i += 1
        else:
            match = True
    if date_format is not None:
        try:
            date_dt = dt.datetime.strptime(date_str, date_format)
        except Exception as e:
            print(f"{date_str} has more complex format than found ({date_format})")
            raise NotImplementedError(e.with_traceback)
    else:
        raise RuntimeError(f"Cannot understand the format of {date_str}")

    return date_dt


def datetime_pytom(d, t):
    """
    Input
        d   Date as an instance of type datetime.date
        t   Time as an instance of type datetime.time
    Output
        The fractional day count since 0-Jan-0000 (proleptic ISO calendar)
        This is the 'datenum' datatype in matlab
    Notes on day counting
        matlab: day one is 1 Jan 0000
        python: day one is 1 Jan 0001
        hence an increase of 366 days, for year 0 AD was a leap year
    """
    dd = d.toordinal() + 366
    tt = dt.timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
    tt = dt.timedelta.total_seconds(tt) / 86400
    return dd + tt


def datetime_mtopy(datenum):
    """
    Input
        The fractional day count according to datenum datatype in matlab
    Output
        The date and time as a instance of type datetime in python
    Notes on day counting
        matlab: day one is 1 Jan 0000
        python: day one is 1 Jan 0001
        hence a reduction of 366 days, for year 0 AD was a leap year
    """
    ii = dt.datetime.fromordinal(int(datenum) - 366)
    ff = dt.timedelta(days=datenum % 1)
    return ii + ff

def datetime_to_raymetrics_string(dt):
    # Extract year, month, and day components from the datetime object
    year = dt.year % 100  # Last two digits of the year
    month = format(dt.month, 'x')  # Convert month to hexadecimal
    day = dt.day

    # Format the date string as "yymdd"
    date_string = f'{year:02d}{month}{day:02d}'
    return date_string



def parse_datetime(date: dt.datetime | dt.date | str | np.datetime64) -> dt.datetime:
    """The general way to cast strigs or datetimes to datetime python type

    Args:
        date (dt.datetime | str): This could be one of the supported formats.
        The recommended one is ISO8601.

    Returns:
        dt.datetime: Python datetime format.
    """

    if isinstance(date, dt.datetime):
        return date
    elif isinstance(date, dt.date):
        return dt.datetime(date.year, date.month, date.day)
    elif isinstance(date, str):
        try:
            return dt.datetime.fromisoformat(date)
        except ValueError:
            return str_to_datetime(date)
    elif isinstance(date, np.datetime64):
        return datetime_np2dt(date)
    else:
        raise ValueError(f"{date} is not a valid date")


# Plot 1:1 line
# axes.plot(axes.xaxis.axes.get_xlim(),axes.xaxis.axes.get_xlim())
def create_linear_interpolator(
    x: np.ndarray, y: np.ndarray
) -> Callable[[np.ndarray], np.ndarray]:
    def interpolator(_x):
        return np.interp(_x, x, y)

    return interpolator


def moving_average(a: np.ndarray, window_size: int = 3) -> np.ndarray:
    # It is better than xarrayDataset.rolling().mean() 
    return sp.ndimage.uniform_filter(a, size=(window_size, 1), mode="nearest")

# sp.ndimage.uniform_filter for xarray DataArrays. The window size should be applied to the dimension 'dim'
def moving_average_xarray(a: xr.DataArray, dim: str | tuple[str, str] = 'range', window_size: int | tuple[int, int] = 3) -> xr.DataArray:

    if isinstance(dim, str):
        if isinstance(window_size, int):
            size = (window_size, 1) if dim == 'range' else (1, window_size)    
        else:
            raise ValueError(f"window_size must be an integer if dim is a string")
    elif isinstance(dim, tuple):        
        if isinstance(window_size, tuple):
            dim=np.array(dim)
            window_size=np.array(window_size)                        
            size = (window_size[dim=='range'].item(), window_size[dim=='time'].item())
        else:
            raise ValueError(f"window_size must be a list if dim is a list")        

    return xr.DataArray(sp.ndimage.uniform_filter(a, size=size, mode="nearest"), dims = a.dims, coords = a.coords)
    

def leapyear(year: int) -> bool:
    """
    determines whether a given year is a leap year or not
    True iff the year is a leap year
    """
    return year % 4 == 0 and year % 100 != 0 or year % 400 == 0


def zellerGregorian(d: int, m: int, a: int) -> str:
    """determines the day of the week corresponding to a temporary
    milestone according to the Gregorian calendar
    """
    # cfr. https://en.wikipedia.org/wiki/Zeller%27s_congruence
    # cfr. https://es.wikipedia.org/wiki/Congruencia_de_Zeller
    if m in [1, 2]:
        m += 12
        a -= 1
    q, m, a = d, m, a
    J, K = divmod(a, 100)
    c1, r1 = divmod((m + 1) * 26, 10)
    c2, r2 = divmod(K, 4)
    c3, r3 = divmod(J, 4)
    r = (q + c1 + K + c2 + c3 - 2 * J) % 7
    return ["sat", "sun", "mon", "tue", "wed", "thu", "fri"][r]


def zellerJulian(d: int, m: int, a: int) -> str:
    """determines the day of the week corresponding to a temporary
    milestone according to the Julian calendar
    """
    # cfr https://www.timeanddate.com/calendar/?country=23
    if m in [1, 2]:
        m += 12
        a -= 1
    q, m, a = d, m, a
    J, K = divmod(a, 100)
    c1, r1 = divmod((m + 1) * 26, 10)
    c2, r2 = divmod(K, 4)
    r = (q + c1 + K + c2 + 5 - J) % 7
    return ["sat", "sun", "mon", "tue", "wed", "thu", "fri"][r]


def weekDay(s: str) -> bool:
    weekValue = {
        "sat": False,
        "sun": False,
        "mon": True,
        "tue": True,
        "wed": True,
        "thu": True,
        "fri": True,
    }
    return weekValue[s]


def holyday(s: str) -> bool:
    """
    holyday('dd/mm/aaaa') gives 'True' iff dd/mm/ isn't in any list
    """
    # Dates given by: https://www.enforex.com/espanol/fiesta-espana.html
    holydaysNat = [
        "01/01",
        "06/01",
        "01/05",
        "15/08",
        "12/10",
        "01/11",
        "06/12",
        "07/12",
        "08/12",
        "25/12",
    ]
    holydaysReg = ["28/02"]
    holydaysLoc = ["02/11", "28/02", "15/09"]
    t = "/".join(s.split("/")[:-1])
    return not (t in holydaysNat or t in holydaysReg or t in holydaysLoc)


def yearHolyday(s: str) -> bool:
    """
    We must have, and thus trust that it is, the "d/m/a" format.
    warning('dd/mm/aa') gives True sii dd/mm/aa is weekend day
    or holiday
    """
    yearHolyday = ["09/04/2020", "10/04/2020", "11/06/2020"]
    return not (s in yearHolyday)


def warning_not_working_date(date: str) -> bool:
    """
    We must have, and thus trust that it is, the "dd/mm/aaaa" format.
    warning('dd/mm/aaaa') gives True sii dd/mm/aa is weekend day
    or holiday
    """
    d, m, a = date.split("/")
    d, m, a = int(d), int(m), int(a)
    lyr = d == 29 and leapyear(a) or d != 29
    return (
        lyr
        and weekDay(zellerGregorian(d, m, a))
        and holyday(date)
        and yearHolyday(date)
    )

def linear_fit(x: np.ndarray, y: np.ndarray) -> dict:
    """Calculate the Durbin-Watson statistic for a given set of residuals.

    Args:
        x (np.ndarray): The independent variable.
        y (np.ndarray): The dependent variable.

    Returns:
        float: The Durbin-Watson statistic.
    """
    # Add a constant for the intercept term
    X = sm.add_constant(x)

    # Fit a linear regression model
    model = sm.OLS(y, X).fit()

    statistics = {}
    # Calculate residuals
    statistics["residuals"] = model.resid
    statistics["parameters"] = model.params  # Index 1 corresponds to the coefficient of x
    statistics["standard_deviation_parameters"] = model.bse  # Index 1 corresponds to the standard error of the coefficient of x
    statistics["msre"] = np.sqrt(np.mean(model.resid ** 2))

    #the Anderson–Darling test coefficient A of the Residual
    statistics["anderson"] = anderson(statistics["residuals"], dist='norm')
    # Perform Durbin-Watson test
    statistics["durbin_watson"] = durbin_watson(statistics["residuals"])
    return statistics


import os
from pathlib import Path
from pdb import set_trace
import sys
import glob
import json
import importlib
import numpy as np
import xarray as xr
import datetime as dt
from loguru import logger
from multiprocessing import Pool

from atmospheric_lidar.scripts import licel2scc

from gfatpy.utils.io import read_yaml

"""
Functions for deriving SCC info directly from measurement folder/files
(Currently Not Used)
"""

def get_scc_code_from_measurement_folder(meas_folder, campaign):
    """
    get scc_code from measurement folder.
    Para todas las scc posibles (2) de la campaña, se busca si los canales que las definen existen en la medida.
    En caso de que existan las 2, la scc correcta es la que tiene más canales.
    Input:
    - measurement folder
    - campaign name
    """

    assert isinstance(meas_folder, str), "meas_folder must be String Type"
    assert isinstance(campaign, str), "campaign must be String Type"

    if campaign is not None:
        campaign_info, campaign_scc_fn = get_campaign_info(campaign)
        scc_cfg = campaign_info.scc_cfg
        scc_codes = campaign_info.scc_codes
        CustomLidarMeasurement = licel2scc.create_custom_class(
            campaign_scc_fn, use_id_as_name=True
        )
        rm_files = glob.glob(os.path.join(meas_folder, "R*"))
        if len(rm_files) > 0:
            rm_file = [rm_files[0]]
            measurement = CustomLidarMeasurement(rm_file)
            channels_in_rm = list(measurement.channels.keys())
            sccs = len(scc_codes)
            exist_scc = [False] * sccs
            channels_in_scc = [0] * sccs
            for i, i_scc in enumerate(scc_cfg):
                exist_scc[i] = all(
                    j in channels_in_rm for j in scc_cfg[i_scc]["channels"]
                )
                channels_in_scc[i] = len(scc_cfg[i_scc]["channels"])
            scc_code = [b for a, b in zip(exist_scc, scc_codes) if a]
            if len(scc_code) == 0:  # no estan los scc posibles en la medida
                scc_code = None
            elif len(scc_code) == 1:  # hay 1.
                scc_code = int(scc_code[0])
            elif (
                len(scc_code) == 2
            ):  # los dos son posibles. elegimos el que tiene mas canales
                max_chan = channels_in_scc == np.max(channels_in_scc)
                scc_code = [b for a, b in zip(max_chan, scc_codes) if a]
                scc_code = int(scc_code[0])
        else:
            scc_code = None
    else:
        scc_code = None
    return scc_code


def get_campaign_info(campaign, scc_config_directory=None):
    """ """
    # TODO: make scc_config_directory optional ¿?. Enlazar con crear archivo de configuracion
    try:
        if scc_config_directory is None:
            scc_config_directory = os.path.join(
                os.path.abspath(__file__), "scc_configFiles"
            )
        if campaign == "covid":  # there is only one campaign so far.
            campaign_scc_fn = "scc_channels_covid19.py"
            campaign_scc_fn = os.path.join(scc_config_directory, campaign_scc_fn)
            campaign_info = import_campaign_scc(campaign_scc_fn)
            campaign_info.scc_codes = [*campaign_info.scc_cfg]
        else:
            campaign_info = None
            campaign_scc_fn = None
    except:
        campaign_info = None
        campaign_scc_fn = None

    return campaign_info, campaign_scc_fn


def import_campaign_scc(campaign_scc_fn):
    """ """
    # TODO: darle una vuelta
    try:
        sys.path.append(os.path.dirname(campaign_scc_fn))
        campaign_scc = importlib.import_module(
            os.path.splitext(os.path.basename(campaign_scc_fn))[0]
        )
    except:
        raise (f"ERROR. importing scc-channels info from {campaign_scc_fn}")
    return campaign_scc

""" Handling Exceedance of Execution Time.
    Inspired in: https://stackoverflow.com/questions/51712256/how-to-skip-to-the-next-input-if-time-is-out-in-python """
# Maximum Allowed Execution Time (seconds)
# Class for timeout exception
class TimeoutException(Exception):
    pass


# Handler function to be called when SIGALRM is received
def sigalrm_handler(signum, frame):
    # We get signal!
    raise TimeoutException()


""" get date from raw lidar files name """
def date_from_filename(filelist):
    """
    It takes the date from the file name of licel files.
    Parameters
    ----------
    filelist: list, str
        list of licel-formatted files

    Returns
    -------
    datelist: list, datetime
        list of datetimes for each file in input list
    """

    datelist = []
    if filelist:
        for _file in filelist:
            body = _file.split(".")[0]
            tail = _file.split(".")[1]
            year = int(body[-7:-5]) + 2000
            month = body[-5:-4]
            try:
                month = int(month)
            except Exception as e:
                if month == "A":
                    month = 10
                elif month == "B":
                    month = 11
                elif month == "C":
                    month = 12
            day = int(body[-4:-2])
            hour = int(body[-2:])
            minute = int(tail[0:2])
            # print('from body %s: the date %s-%s-%s' % (body, year, month, day))
            cdate = dt.datetime(year, month, day, hour, minute)
            datelist.append(cdate)
    else:
        print("Filelist is empty.")
        datelist = None
    return datelist


def getTP(filepath):
    """
    Get temperature and pressure from header of a licel-formatted binary file.
    Inputs:
    - filepath: path of a licel-formatted binary file (str)
    Output:
    - temperature: temperature in celsius (float).
    - pressure: pressure in hPa (float).
    """
    # This code should evolve to read the whole header, not only temperature and pressure.
    if os.path.isfile(filepath):
        with open(filepath, mode="rb") as f:  #
            filename = f.readline().decode("utf-8").rstrip()[1:]
            second_line = f.readline().decode("utf-8")
            f.close()
        second_line_list = second_line.split(" ")
        if len(second_line_list) == 14:
            temperature = float(second_line_list[12].rstrip())
            pressure = float(second_line_list[13].rstrip())
        else:
            logger.warning("Cannot find temperature, pressure values. set to None")
            temperature = None
            pressure = None
    else:
        logger.warning("File not found.")
        temperature = None
        pressure = None
    return temperature, pressure


def apply_pc_peak_correction(filelist, scc_pc_channels):
    """
    Correction of the PC peaks in the PC channels caused by PMT degradation.

    Parameters
    ----------
    filelist: list(str)
        File list (e.g., /c/*.nc') (list)

    Returns
    -------
    outputlist: list(str)
        NetCDF file [file]
    """
    outputlist = list()
    threshold = 1000

    # scc_pc_channels = [1047, 1048, 1090, 1093, 1094]  # TODO: esto aqui a fuego ...
    if np.logical_and(len(filelist) > 0, len(scc_pc_channels) > 0):
        for file_ in filelist:
            try:
                lxarray = xr.open_dataset(file_)
                output_directory = os.path.dirname(file_)
                filename = os.path.basename(file_)
                for channel_ in scc_pc_channels:
                    idx_channel = np.where(lxarray.channel_ID == channel_)[0]
                    if idx_channel.size > 0:
                        # Call pc_peak_correction from utils_gfat
                        profile_raw = lxarray.Raw_Lidar_Data[:, idx_channel, :].values
                        shape_raw = profile_raw.shape
                        profile = np.squeeze(profile_raw)
                        new_profile = mulhacen_pc_peak_correction(profile)
                        lxarray.Raw_Lidar_Data[:, idx_channel, :] = np.reshape(
                            new_profile.astype("int"), shape_raw
                        )

                        profile_raw = lxarray.Background_Profile[
                            :, idx_channel, :
                        ].values
                        shape_raw = profile_raw.shape
                        profile = np.squeeze(profile_raw)
                        new_profile = mulhacen_pc_peak_correction(profile)
                        lxarray.Background_Profile[:, idx_channel, :] = np.reshape(
                            new_profile.astype("int"), shape_raw
                        )

                # save corrected data in the same file
                auxfilepath = os.path.join(output_directory, "aux")
                lxarray.to_netcdf(path=auxfilepath, mode="w")
                os.remove(file_)
                os.rename(auxfilepath, file_)
            except Exception as e:
                logger.warning(str(e))
                logger.warning("PC peak correction not performed")
            outputlist.append(file_)
    else:
        print("Files not found.")
    return outputlist


def get_info_from_measurement_file(mea_fn, scc_config_fn):
    """
    From a R File, extract information

    Parameters
    ----------
    mea_fn: str
        full path of measurement file
    scc_config_fn: str
        py file of scc configuration

    Returns
    -------
    time_ini_i: datetime.datetime
        initial time
    time_end_i: datetime.datetime
        end time
    channels_in_rm: collections.OrderedDict
        channels info

    """
    CustomLidarMeasurement = licel2scc.create_custom_class(
        scc_config_fn, use_id_as_name=True
    )
    mea = CustomLidarMeasurement([mea_fn])  # MUY LENTO
    time_ini_i = mea.info["start_time"].replace(tzinfo=None)
    time_end_i = mea.info["stop_time"].replace(tzinfo=None)
    # channels: Object LicelChannel (atmospheric_lidar/licel.py, L516)
    channels_in_rm = mea.channels
    del CustomLidarMeasurement, mea
    return time_ini_i, time_end_i, channels_in_rm


def get_info_from_measurement_files(meas_files, scc_config_fn):
    """
    given a list of measurement files, information about:
        - scc configuration
        - time lapse of measurement is taken
    a scc_config.py file is needed to read file contents. [mhc_parameters_scc_xxx.py]

    Parameters
    ----------
    meas_files : [type]
        [description]
    scc_config_fn : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    try:
        # Parallelization: reduces near 1 minute when normal takes 2 minutes.
        with Pool(os.cpu_count()) as pool:
            x = [
                (
                    mea_fn,
                    scc_config_fn,
                )
                for mea_fn in meas_files
            ]
            res = np.array(pool.starmap(get_info_from_measurement_file, x))
        times_ini = [x[0] for x in res]
        times_end = [x[1] for x in res]
        channels_in_rm = [x[2] for x in res][0]
        time_start = times_ini[0].replace(second=0)
        time_end = times_end[-1] + dt.timedelta(minutes=1)
        time_end = time_end.replace(second=0)

        """ 
        # Non-parallel
        t1 = time.time()
        times_ini_s = []
        times_end_s = []
        for i, m_file in enumerate(meas_files):
            # times ini, end
            time_ini_i, time_end_i, channels_in_rm = get_info_from_measurement_file(scc_config_fn, m_file)
            if i == 0:
                # time of first measurement starts
                time_start = time_ini_i.replace(second=0)
            if i == len(meas_files) - 1:
                # time of last measurement ends
                time_end = time_end_i + dt.timedelta(minutes=1)
                time_end = time_end.replace(second=0)
            times_ini_s.append(time_ini_i)
            times_end_s.append(time_end_i)
        print(time.time() - t1)
        """
    except Exception as e:
        logger.error(str(e))
        logger.error("Measurement files not read")
        return

    return times_ini, times_end, time_start, time_end, channels_in_rm


def get_scc_config(scc_config_fn):
    """

    Parameters
    ----------
    scc_config_fn: str
        scc configuration file

    Returns
    -------
    scc_config_dict: dict
        Dictionary with scc configuration info from scc config file


    """
    try:
        sys.path.append(os.path.dirname(scc_config_fn))
        module = importlib.import_module(
            os.path.splitext(os.path.basename(scc_config_fn))[0]
        )
        scc_config_dict = dict()
        scc_config_dict["general_parameters"] = module.general_parameters
        scc_config_dict["channel_parameters"] = module.channel_parameters
        scc_config_dict["channels"] = [*scc_config_dict["channel_parameters"]]
    except:
        logger.warning("ERROR. importing scc parameteres from %s" % scc_config_fn)
        scc_config_dict = None

    return scc_config_dict


def get_campaign_config(
    campaign_cfg_fn=None,
    scc_config_id=None,
    hour_ini=None,
    hour_end=None,
    hour_resolution=None,
    timestamp=0,
    slot_name_type=0,
):
    """
    Get Campaign Info from file into a dictionary
    If not campaign_cfg_fn is given, a campaign_cfg is built, using,
    if scc_config_id is given

    Campaign config file is a json with different configurations as keys

    Parameters
    ----------
    campaign_cfg_fn : str
        campaign config file. Default: GFATserver
    scc_config_id: int
        scc lidar configuration number
    hour_ini: float
    hour_end: float
    hour_resolution: float
    timestamp: int
        0: timestamp at beginning of interval
        1: timestamp at center of interval
    slot_name_type: int
        0: earlinet: YYYYMMDD+station+slot_number.
        1: scc campaigns: YYYYMMDD+station+HHMM.
        2: alternative: YYYYMMDD+station+HHMM+_scc

    Returns
    -------
    scc_campaign_cfg: dict

    """

    if campaign_cfg_fn is None:
        campaign_cfg_fn = "GFATserver"

    if campaign_cfg_fn == "GFATserver":  # Default Campaign. Dictionary as template
        scc_campaign_cfg = {
            "name": "operational",
            "lidar_config": {
                "operational": {
                    "scc": scc_config_id,
                    "hour_ini": hour_ini,
                    "hour_end": hour_end,
                    "hour_res": hour_resolution,
                    "timestamp": timestamp,
                    "slot_name_type": slot_name_type,
                }
            },
        }
    else:  # If Campaign File is given
        if os.path.isfile(campaign_cfg_fn):
            with open(campaign_cfg_fn) as f:
                scc_campaign_cfg = json.load(f)
        else:
            logger.error(
                "Campaign File %s Does Not Exist. Exit program" % campaign_cfg_fn
            )
            sys.exit()
    return scc_campaign_cfg


def check_scc_output_inlocal(scc_output_slot_dn):
    """
    Check if Products have been downloaded from SCC server for a given slot output directory

    Parameters
    ----------
    scc_output_slot_dn: str
        full path local directory of scc slot
    Returns
    -------
    scc_output_inlocal: bool
        False if something has not been downloaded from SCC server
    """

    expected_dns = [
        "hirelpp",
        "cloudmask",
        "scc_preprocessed",
        "scc_optical",
        "scc_plots",
    ]
    exist_dns = [
        os.path.isdir(os.path.join(scc_output_slot_dn, i)) for i in expected_dns
    ]
    if not all(exist_dns):
        scc_output_inlocal = False
    else:
        if len(glob.glob(os.path.join(scc_output_slot_dn, "scc_optical"))) == 0:
            scc_output_inlocal = False
        if len(glob.glob(os.path.join(scc_output_slot_dn, "scc_plots"))) == 0:
            scc_output_inlocal = False
        else:
            scc_output_inlocal = True

    return scc_output_inlocal



"""
Config File for atmospheric_lidar.scripts.licel2scc.py
licel2scc.create_custom_class(this_file) -> CustomLidarMeasurement
"""

""" general parameters """
general_parameters = {
    'System': '\'MULHACEN\'',
    'Laser_Pointing_Angle': 0,
    'Molecular_Calc': 0,
#     'Latitude_degrees_north': 37.164,
#     'Longitude_degrees_east': -3.605,
#     'Altitude_meter_asl': 680.0,
    'Call sign': "gr"
    }

"""  create dictionary channel_parameters """
# licel channels to be used in scc config linked to scc channels id
licel_channel_id = {
    'BT0': 1091,
    'BT1': 1092,
    'BT2': 1049,
    'BT3': 1045,
    'BC0': 1094,
    'BC1': 1093,
    'BC2': 1090,
    'BC4': 1047,
}
channel_parameters = dict()
for k, v in licel_channel_id.items():
    channel_parameters[k] = {
        'channel_ID': v,
        'Background_Low': 75000,
        'Background_High': 105000,
        'LR_Input': 1,
        }

# Old fashioned
#channel_parameters = {
#    'BT0': {'channel_ID': 1091,
#            'Background_Low': 75000,
#            'Background_High': 105000,
#            'LR_Input': 1, },
#    'BT1': {'channel_ID': 1092,
#            'Background_Low': 75000,
#            'Background_High': 105000,
#            'LR_Input': 1, },
#    'BT2': {'channel_ID': 1049,
#             'Background_Low': 75000,
#             'Background_High': 105000,
#             'LR_Input': 1, },
#    'BT3': {'channel_ID': 1045,
#             'Background_Low': 75000,
#             'Background_High': 105000,
#             'LR_Input': 1, },               
#    'BC0': {'channel_ID': 1094,
#            'Background_Low': 75000,
#            'Background_High': 105000,
#            'LR_Input': 1, },               
#    'BC1': {'channel_ID': 1093,
#            'Background_Low': 75000,
#            'Background_High': 105000,
#            'LR_Input': 1, },               
#    'BC2': {'channel_ID': 1090,
#            'Background_Low': 75000,
#            'Background_High': 105000,
#            'LR_Input': 1, },               
#    'BC4': {'channel_ID': 1047,
#            'Background_Low': 75000,
#            'Background_High': 105000,
#            'LR_Input': 1, },               
#    }



from pathlib import Path
from pdb import set_trace
from typing import Any, Tuple, Union
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from scipy import integrate
from scipy.integrate import cumulative_trapezoid as cumtrapz

from loguru import logger

from gfatpy.atmo.atmo import transmittance
from gfatpy.lidar.utils.types import ParamsDict
from gfatpy.lidar.utils.utils import refill_overlap, signal_to_rcs
from gfatpy.utils.plot import color_list


def klett_rcs(
    rcs_profile: np.ndarray[Any, np.dtype[np.float64]],
    range_profile: np.ndarray[Any, np.dtype[np.float64]],
    beta_mol_profile: np.ndarray[Any, np.dtype[np.float64]],
    lr_part: float | np.ndarray[Any, np.dtype[np.float64]] = 45.,
    lr_mol: float = 8 * np.pi / 3,
    reference: Tuple[float, float] = (8000, 8500),
    beta_aer_ref: float = 0,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """Calculate aerosol backscattering using Classical Klett algorithm verified with Fernald, F. G.: Appl. Opt., 23, 652-653, 1984.

    Args:
        rcs_profile (np.ndarray): 1D signal profile.
        range_profile (np.ndarray): 1D range profile with the same shape as rcs_profile.
        beta_mol_profile (np.ndarray): 1D array containing molecular backscatter values.
        lr_mol (float): Molecular lidar ratio (default value based on Rayleigh scattering).
        lr_part (float, optional): Aerosol lidar ratio (default is 45 sr).
        reference (Tuple[float, float], optional): Range interval (ymin and ymax) for reference calculation. Defaults to (8000, 8500).
        beta_aer_ref (float, optional): Aerosol backscatter at reference range (ymin and ymax). Defaults to 0.

    Returns:
        np.ndarray: Aerosol-particle backscattering profile.
    """

    if isinstance(lr_part, float):
        lr_part = np.full(len(range_profile), lr_part)
    if isinstance(lr_mol, float):
        lr_mol = np.full(len(range_profile), lr_mol) # type: ignore

    ymin, ymax = reference

    particle_beta = np.zeros(len(range_profile))

    ytop = np.abs(range_profile - ymax).argmin()

    range_resolution = np.median(np.diff(range_profile)).astype(float)

    idx_ref = np.logical_and(range_profile >= ymin, range_profile <= ymax)

    if not idx_ref.any():
        raise ValueError("Range [ymin, ymax] out of rcs size.")

    calib = np.nanmean(
        rcs_profile[idx_ref] / (beta_mol_profile[idx_ref] + beta_aer_ref)
    )

    # from Correct(ed) Klett–Fernald algorithm for elastic aerosol backscatter retrievals: a sensitivity analysis
    # Johannes Speidel* AND Hannes Vogelmann
    # https://doi.org/10.1364/AO.465944
    # Eq. 10
    # Reminder: BR = lr_mol, BP = lr_part

    integral_in_Y = np.flip(cumtrapz( np.flip((lr_mol[:ytop] - lr_part[:ytop]) * beta_mol_profile[:ytop]), dx=range_resolution, initial=0) ) # type: ignore
    exp_Y = np.exp( -2 * integral_in_Y)

    integral_in_particle_beta = np.flip(cumtrapz(np.flip(lr_part[:ytop] * rcs_profile[:ytop] * exp_Y), dx=range_resolution, initial=0)) # type: ignore

    total_beta = (rcs_profile[:ytop] * exp_Y) / (calib + 2 * integral_in_particle_beta)

    particle_beta[:ytop] = total_beta - beta_mol_profile[:ytop]

    return particle_beta


def quasi_beta(
    rcs_profile: np.ndarray[Any, np.dtype[np.float64]],
    calibration_factor: float,
    range_profile: np.ndarray[Any, np.dtype[np.float64]],
    params: dict | ParamsDict,
    lr_part: float | np.ndarray = 45.0,    
    full_overlap_height: float = 1000.0,    
    debug: bool = False,
) -> np.ndarray:

    #calculate beta attenuated 
    att_beta = rcs_profile / calibration_factor 
    
    star_beta = att_beta / transmittance(params["molecular_alpha"], range_profile)**2 - params["molecular_beta"]

    star_alpha = star_beta * lr_part

    #refill overlap
    star_alpha = refill_overlap(star_alpha, range_profile, full_overlap_height) # type: ignore

    # T2 = np.ones(len(range_profile))
    # for i in range(1, len(range_profile)):
    #     T2[i] = T2[i-1] * np.exp(-2 * (params["molecular_alpha"][i-1] + star_alpha[i-1]) * (range_profile[i] - range_profile[i-1]))

    # quasi_beta1 = att_beta / T2  - params["molecular_beta"]
    quasi_beta = att_beta / transmittance(params["molecular_alpha"] + star_alpha, range_profile)**2 - params["molecular_beta"]

    quasi_alpha = quasi_beta * lr_part    

    if debug:
        fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(10, 5), sharey=True)
        #Axis 1: plot RCS
        ax[0].plot(att_beta, range_profile, lw=2, label="attenuated beta")
        ax[0].set_xscale("log")
        ax[0].set_ylim(0, 9000)
        ax[0].set_xlabel("attenuated beta, [a.u.]")
        ax[0].legend()    
        #Axis 2: plot calibration
        ax[1].plot(1e6*quasi_beta, range_profile, lw=2, label="quasi beta")
        ax[1].plot(1e4*quasi_alpha, range_profile, lw=2, ls = '--', label="star alpha")
        ax[1].set_xscale("linear")
        ax[1].set_ylim(0, 9000)
        ax[1].set_xlabel("quasi beta, [a.u.]")
        ax[1].legend()
        fig.savefig(Path().cwd() / 'quasi_beta.png')

    return quasi_beta

def iterative_beta(
    rcs_profile: np.ndarray,
    calibration_factor: float,
    range_profile: np.ndarray,
    params: dict | ParamsDict,
    lr_part: float | np.ndarray = 45.0,    
    full_overlap_height: float = 1000.0,
    free_troposphere_height: float = 5000.0,    
    debug: bool = False,
    iterations: int = 10,
    tolerance: float = 0.01
) -> np.ndarray:

    alpha_part_previous = np.zeros(len(range_profile))
    beta_part = np.zeros((iterations, len(range_profile)))
    backscattering_ratio = np.zeros((iterations, len(range_profile)))
    relative_diff_beta_previous = 1.0
    beta_part_previous = params["molecular_beta"]
    colors = color_list(iterations)
    resolution = np.median(np.diff(range_profile)).astype(float)

    for idx in range(iterations):
        #Molecular atmosphere
        signal_mol = calibration_factor * params["molecular_beta"] * transmittance(params["molecular_alpha"] + alpha_part_previous, range_profile)**2 / range_profile**2
        rcs_mol = signal_to_rcs(signal_mol, range_profile)
        
        #calculate backscattering ratio
        R = rcs_profile / rcs_mol - 1

        beta_part_current = R * params["molecular_beta"]
        
        if debug:
            beta_part[idx,:] = beta_part_current
            backscattering_ratio[idx,:] = R            

        integral_beta_previous = np.trapz(beta_part_previous[range_profile < free_troposphere_height], dx=resolution)
        integral_beta_current = np.trapz(beta_part_current[range_profile < free_troposphere_height], dx=resolution)

        relative_diff_beta_current = np.abs(integral_beta_current - integral_beta_previous) / integral_beta_previous

        if np.abs(relative_diff_beta_current) < tolerance:            
            break
        else:
            beta_part_previous = beta_part_current
            relative_diff_beta_previous = relative_diff_beta_current            
            alpha_part_previous = refill_overlap(beta_part_current * lr_part, range_profile, full_overlap_height)

    #raise if no convergence
    # if idx == iterations - 1:
    #     raise ValueError("No convergence in iterative beta retrieval.")

    if debug:
        fig, ax = plt.subplots(ncols=4, nrows=1, figsize=(10, 5), sharey=True)
        #Axis 1: plot rcs mol
        ax[0].plot(rcs_mol, range_profile, lw=2, label="mol")
        ax[0].plot(rcs_profile, range_profile, lw=2, label="part")
        ax[0].set_xscale("log")
        ax[0].set_ylim(0, 9000)
        ax[0].set_xlabel("RCS, [a.u.]")
        ax[0].legend()
        #Axis 4: plot backscattering ratio
        for i in range(iterations):
            ax[1].plot(backscattering_ratio[i,:], range_profile, lw=2, color=colors[i], label=f"{i}")        
        ax[1].set_xscale("linear")
        ax[1].set_ylim(0, 9000)
        ax[1].set_xlabel("R, [a.u.]")
        ax[1].legend(fontsize=5)
        #Axis 3: plot beta part
        if "particle_beta" in params:
            ax[2].plot(1e6*params["particle_beta"], range_profile, lw=2, label="synthetic")
        for i in range(iterations):
            ax[2].plot(1e6*beta_part[i,:], range_profile, lw=2, color=colors[i], label=f"{i}")        
        ax[2].set_xscale("linear")
        ax[2].set_ylim(0, 9000)
        ax[2].set_xlabel(r"$\beta$, [a.u.]")
        ax[2].legend(fontsize=5)
        #Axis 4: plot relative diff of beta
        for i in range(iterations):
            ax[3].plot(100*(beta_part[i,range_profile < free_troposphere_height] - params['particle_beta'][range_profile < free_troposphere_height])/params['particle_beta'][range_profile < free_troposphere_height], range_profile[range_profile < free_troposphere_height], lw=2, color=colors[i], label=f"{i}")
        ax[3].set_xscale("linear")
        ax[3].set_ylim(0, 9000)
        ax[3].set_xlabel(r"$\varepsilon_{\beta_{p}}$, [%]")
        ax[3].legend(fontsize=5)

        fig.savefig(Path(__file__).parent.parent.parent.parent / 'tests' / 'figures' / 'iterative_beta.png')        
    
    return beta_part_current
    

def klett_likely_bins(
    rcs_profile: np.ndarray[Any, np.dtype[np.float64]],
    att_mol_beta: np.ndarray[Any, np.dtype[np.float64]],
    heights: np.ndarray[Any, np.dtype[np.float64]],
    min_height: float = 1000,
    max_height: float = 1010,
    window_size: int = 50,
    step: int = 1,
):
    window_size // 2
    i_bin, e_bin = np.searchsorted(heights, [min_height, max_height])

    for i in np.arange(i_bin, e_bin + 1):
        rcs_profile / rcs_profile

    return rcs_profile



def find_lidar_ratio(
    rcs: np.ndarray[Any, np.dtype[np.float64]],
    height: np.ndarray[Any, np.dtype[np.float64]],
    beta_mol: np.ndarray[Any, np.dtype[np.float64]],
    lr_mol: float,
    reference_aod: float,
    mininum_height: float = 0,
    lr_initial: float = 50,
    lr_resol: float = 1,
    max_iterations: int = 100,
    rel_diff_aod_percentage_threshold: float = 1,
    debugging: bool = False,
    klett_reference: Tuple[float, float] = (7000, 8000),
) -> Tuple[float, float | None, bool]:
    """Iterative process to find the lidar ratio (lr) that minimizes the difference between the measured and the calculated aerosol optical depth (aod).

    Args:
        rcs (np.ndarray[Any, np.dtype[np.float64]]): Range Corrected Signal
        height (np.ndarray[Any, np.dtype[np.float64]]): Range profile
        beta_mol (np.ndarray[Any, np.dtype[np.float64]]): Molecular backscattering coefficient profile
        lr_mol (float): Molecular lidar ratio
        reference_aod (float): Reference aerosol optical depth
        mininum_height (float, optional): Fullover height. Defaults to 0.
        lr_initial (float, optional): _description_. Defaults to 50.
        lr_resol (float, optional): _description_. Defaults to 1.
        max_iterations (int, optional): _description_. Defaults to 100.
        rel_diff_aod_percentage_threshold (float, optional): _description_. Defaults to 1.
        debugging (bool, optional): _description_. Defaults to False.
        klett_reference (Tuple[float, float], optional): _description_. Defaults to (7000, 8000).

    Returns:
        Tuple[float, float | None, bool]: _description_
    """

    # Calculate range resolution
    range_resolution = np.median(np.diff(height)).item()

    # Initialize loop
    lr_, iter_, run, success = lr_initial, 0, True, False
    rel_diff_aod = None

    while run:
        iter_ = iter_ + 1

        # Calculate aerosol backscatter
        beta_ = klett_rcs(
            rcs, height, beta_mol, lr_part=lr_, lr_mol=lr_mol, reference=klett_reference
        )

        # Refill beta profile from minimum height to surface to avoid overlap influence
        beta_ = refill_overlap(beta_, height, fulloverlap_height=mininum_height)

        # Calculate aerosol optical depth
        aod_ = integrate.simps(beta_ * lr_, dx=range_resolution)

        # Calculate relative difference between measured and calculated aod
        rel_diff_aod = 100 * (aod_ - reference_aod) / reference_aod

        if debugging:
            print(
                "lidar_ratio: %.1f | lidar_aod: %.3f| reference_aod: %.3f | relative_difference: %.1f%%"
                % (lr_, aod_, reference_aod, rel_diff_aod)
            )

        # Check convergence
        if np.abs(rel_diff_aod) > rel_diff_aod_percentage_threshold:
            if rel_diff_aod > 0:
                if lr_ < 20:
                    run = False
                    print("No convergence. LR goes too low.")
                else:
                    lr_ = lr_ - 1
            else:
                if lr_ > 150:
                    run = False
                    print("No convergence. LR goes too high.")
                else:
                    lr_ = lr_ + 1
        else:
            print("LR found: %f" % lr_)
            run = False
            success = True

        # Check maximum number of iterations
        if iter_ == max_iterations:
            run = False
            print("No convergence. Too many iterations.")

    return lr_, rel_diff_aod, success

def get_calibration_factor(P_elastic: np.ndarray,
                           range_profile: np.ndarray,
                           params: dict | pd.DataFrame,
                           beta_part_profile: np.ndarray,
                           lr_part: float | np.ndarray,                           
                           range_to_average: Tuple[float, float],
                           full_overlap_height: float = 1000.,
                           backscattering_ratio_threshold: float = 1.1,
                           debug: bool = False
                           ) -> Union[Tuple[float, float], Tuple[np.ndarray, np.ndarray]]:

    #fill overlap
    beta_part_profile = refill_overlap(beta_part_profile, range_profile, full_overlap_height)
       
    if isinstance(lr_part, float):
        lr_part = np.full(len(range_profile), lr_part)
    
    #Retrieve inverse squared transmittance 
    T2_elastic = 1/transmittance(params["molecular_alpha"] + beta_part_profile*lr_part, range_profile)**2 # type: ignore
    
    #Get calibration factor
    calibration = signal_to_rcs(P_elastic, range_profile) / (params["molecular_beta"] + beta_part_profile) * T2_elastic

    # Cut heights below overlap height
    calibration = refill_overlap(calibration, range_profile, full_overlap_height, fill_with=np.nan)

    # Filter for backscatter ratio greater than 1.1 and only lower layers
    bakcscattering_ratio = (beta_part_profile + params['molecular_beta']) / params['molecular_beta']
    calibration[bakcscattering_ratio < backscattering_ratio_threshold] = np.nan

    # Filter for area with stable standard deviation    
    ymin, ymax = range_to_average
    idx_ref = np.logical_and(range_profile >= ymin, range_profile <= ymax)
    mean_calib = np.nanmean(calibration[idx_ref])
    std_calib = np.nanstd(calibration[idx_ref])   

    if debug:
        fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(10, 5), sharey=True)
        #Axis 1: plot RCS
        ax[0].plot(signal_to_rcs(P_elastic, range_profile), range_profile, lw=2, label="RCS")
        ax[0].set_xscale("log")
        # ax[0].set_xlim(1e1, 1e6)
        ax[0].set_ylim(0, 9000)
        ax[0].set_xlabel("RCS, [a.u.]")
        ax[0].legend()
        #Axis 2: plot calibration
        ax[1].plot(calibration, range_profile, lw=2, label="calibration")
        ax[1].set_xscale("linear")
        # ax[1].set_xlim(0., 20.)
        ax[1].set_ylim(0, 9000)
        ax[1].set_xlabel("calibration, [a.u.]")
        ax[1].legend()
        fig.savefig(Path(__file__).parent.parent.parent.parent / 'tests' / 'figures' / 'get_calibration_factor.png')
    
    return mean_calib, std_calib # type: ignore



import numpy as np
import xarray as xr

from gfatpy.lidar.scc.plot.product import SCC_elda


def angstrom_exponent(product1: SCC_elda, product2: SCC_elda) -> xr.DataArray | None:
    """Retrive Angstrom exponent with scc products. 

    Args:
        product1 (SCC_elda): Elda product object
        product2 (SCC_elda): Elda product object.

    Returns:
        xr.DataArray | None
    """
    elda1, elda2 = xr.open_dataset(product1.path), xr.open_dataset(product2.path)
    backscatter1 = elda1.backscatter.sel(wavelength=product1.wavelength)
    backscatter2 = elda2.backscatter.sel(wavelength=product2.wavelength)
    if len(elda1.altitude) > len(elda2.altitude):
        interpolated_backscatter2 = elda2.backscatter.interp(altitude=elda1.backscatter.altitude, method='nearest')
        ratio = interpolated_backscatter2/backscatter1
        positive = np.logical_and(interpolated_backscatter2>0,backscatter1>0)
    else:
        interpolated_backscatter1 = elda1.backscatter.interp(altitude=elda2.backscatter.altitude, method='nearest')
        ratio = backscatter2/interpolated_backscatter1
        positive = np.logical_and(backscatter2>0,interpolated_backscatter1>0)
    
    ae = (-np.log(ratio.where(positive))/np.log(product2.wavelength/product1.wavelength)).squeeze()
    if len(ae.values)==0:
        ae = None
    return ae


import numpy as np
import os

class GlobalInfo:
    filename = ''
    Location = ''
    StartTime = ''
    StopTime = ''
    Height = 0
    Longitude = 0.0
    Latitude = 0.0
    Zenith = 0.0
    Azimuth = 0.0
    Custom = ''
    numShotsL0 = 0
    repRateL0 = 0
    numShotsL1 = 0
    repRateL1 = 0
    numShotsL2 = 0
    repRateL2 = 0
    numDataSets = 0
class dataSet:
    active = 0
    dataType = 0
    laserSource = 0
    numBins = 0 
    laserPolarization = 0
    highVoltage = 0
    binWidth = 0.0
    LaserWavelength = 0.0
    Polarisation = 'o'
    binshift = 0
    binshiftPart = 0
    ADCBits = 0
    numShots = 0
    inputRange = 0.5
    discriminator = 0.003
    descriptor = ''
    rawData  = np.zeros((3,))
    physData = np.zeros((3,))
    
  
    def __init__(self, stringIn):
      self.active = int(stringIn.split()[0]) 
      self.dataType = int(stringIn.split()[1])
      self.laserSource = int(stringIn.split()[2])
      self.numBins     = int(stringIn.split()[3])
      self.laserPolarization = int(stringIn.split()[4])
      self.highVoltage = int(stringIn.split()[5])
      (self.binwidth) = float(stringIn.split()[6])
      self.LaserWavelength = int(stringIn.split()[7].split('.')[0])
      self.binshift = int(stringIn.split()[8])
      self.ADCBits = int(stringIn.split()[12])
      self.numShots = int(stringIn.split()[13])
      if ((self.dataType == 0) or (self.dataType == 2)):
        self.inputRange = float(stringIn.split()[14])
      else: 
        self.discriminator = float(stringIn.split()[14])
      self.descriptor = stringIn.split()[15]
    


class LicelFileReader:
  def __init__(self, filename):
     self.GlobalInfo = GlobalInfo
     self.dataSet = []
     fp = open(filename, 'rb')
     encoding = 'utf-8'
     self.GlobalInfo.filename = str(fp.readline(), encoding).split()[0];
     self.firstline = str(fp.readline(), encoding)
     self.secondline = str(fp.readline(), encoding)
     self.GlobalInfo.Location  = self.firstline.split()[0]
     self.GlobalInfo.StartTime = self.firstline.split()[1] + ' ' + self.firstline.split()[2]
     self.GlobalInfo.StopTime  =  self.firstline.split()[3] + ' ' + self.firstline.split()[4]
     self.GlobalInfo.Height = int(self.firstline.split()[5])
     self.GlobalInfo.Longitude = float(self.firstline.split()[6])
     self.GlobalInfo.Latitude = float(self.firstline.split()[7])
     self.GlobalInfo.Zenith = float(self.firstline.split()[8])
     if (len(self.firstline.split()) > 10): 
        self.GlobalInfo.Azimuth = float(self.firstline.split()[9])
        
     self.GlobalInfo.numShotsL0  = int(self.secondline.split()[0])
     self.GlobalInfo.repRateL0  = int(self.secondline.split()[1])
     self.GlobalInfo.numShotsL1  = int(self.secondline.split()[2])
     self.GlobalInfo.reprateL1  = int(self.secondline.split()[3])
     self.GlobalInfo.numDataSets   = int(self.secondline.split()[4])
     #self.GlobalInfo.numShotsL2  = int(self.secondline.split()[5])
     #self.GlobalInfo.repRateL2  = int(self.secondline.split()[6])
     for i in range(self.GlobalInfo.numDataSets):
       self.varline = str(fp.readline(), encoding)
       self.dataSet.append(dataSet(self.varline))
     fp.readline()
     for i in range(self.GlobalInfo.numDataSets):
       if (i > 0):
        fp.read(2)
       self.dataSet[i].rawData = np.fromfile(fp, dtype=int, count = self.dataSet[i].numBins)
       if ((self.dataSet[i].dataType == 0) and (self.dataSet[i].ADCBits > 1)): #analog
         maxbits = (2 ** self.dataSet[i].ADCBits) - 1
         self.dataSet[i].physData = self.dataSet[i].inputRange * self.dataSet[i].rawData / (self.dataSet[i].numShots if self.dataSet[i].numShots > 0 else 1) /maxbits
       elif (self.dataSet[i].dataType == 1): # pc
         self.dataSet[i].physData = self.dataSet[i].rawData / (self.dataSet[i].numShots if self.dataSet[i].numShots > 0 else 1) * (150 / self.dataSet[i].binwidth)
       elif (self.dataSet[i].dataType == 2): # analog sqr
         maxbits = (2 ** self.dataSet[i].ADCBits) - 1
         n = (self.dataSet[i].numShots if self.dataSet[i].numShots > 0 else 1)
         sq_n_1 = np.sqrt(self.dataSet[i].numShots -1) if (self.dataSet[i].numShots > 1) else 1
         self.dataSet[i].physData = self.dataSet[i].inputRange * self.dataSet[i].rawData /(n * sq_n_1) /maxbits 
       elif (self.dataSet[i].dataType == 3): # pc sqr
         self.dataSet[i].physData = self.dataSet[i].rawData / (self.dataSet[i].numShots if self.dataSet[i].numShots > 0 else 1) * (150 / self.dataSet[i].binwidth) / np.sqrt((self.dataSet[i].numShots -1) if self.dataSet[i].numShots > 1 else 1)
       else :
         self.dataSet[i].physData = self.dataSet[i].rawData / (self.dataSet[i].numShots if self.dataSet[i].numShots > 0 else 1)
     fp.close()









from pathlib import Path

import xarray as xr

from pathlib import Path

from pathlib import Path
from typing import Dict, Union


def ghk_output_reader(filepath: Path) -> Dict[str, Union[float, None]]:
    """It loads in a dictionary the ghk parameters from `filepath`.

    Args:
        filepath (Path): ghk filepath generated for the Volker's software `GHK_0.9.8h_Py3.7.py`.

    Raises:
        FileNotFoundError: filepath not found.
        ValueError: GHK parameters not found in the filepath.

    Returns:
        Dict[str, Union[float, None]]: Dictionary with keys: GR, GT, HR, HT, K1, K2, ..., K7.
    """    
    GHK: Dict[str, Union[float, None]] = {
        "GR": None,
        "GT": None,
        "HR": None,
        "HT": None,
        "K1": None,
        "K2": None,
        "K3": None,
        "K4": None,
        "K5": None,
        "K6": None,
        "K7": None,
    }
    if not filepath.is_file():
        raise FileNotFoundError(f"{filepath} not found.")

    with open(filepath, "r") as f:
        for line in f:
            line = line.strip()
            if line[0:2] == "GR":
                break
        line = f.readline().replace(" ", "").split(",")
        try:
            (
                GHK["GR"],
                GHK["GT"],
                GHK["HR"],
                GHK["HT"],
                GHK["K1"],
                GHK["K2"],
                GHK["K3"],
                GHK["K4"],
                GHK["K5"],
                GHK["K6"],
                GHK["K7"],
            ) = [float(val) if val else None for val in line]
        except (ValueError, TypeError, IndexError):
            raise ValueError("Unexpected line format encountered.")

    return GHK





import os

import numpy as np
import numpy as np
import xarray as xr
from datetime import datetime, timedelta
import matplotlib as mpl
import matplotlib.pyplot as plt
from loguru import logger


def plotQuicklook(data: xr.Dataset, plt_conf: xr.Dataset, saveImageFlag: bool = False):

    logger.warning("to be deprecated. Use cloudnetpy instead")
    """
    Inputs:
    - data: from categorice_reader()
    - plt_conf: plot configuration dictionary as follows:
        plt_conf =  {
            'mainpath': "Y:\\datos\\CLOUDNET\\juelich\\quicklooks",
            'coeff': COEFF,
            'gapsize': HOLE_SIZE,
            'dpi': dpi,
            'fig_size': (16,5),
            'font_size': 16,
            'y_min': 0,
            'y_max': range_limit,
            'rcs_error_threshold':1.0, }
    - saveImageFlag [Boolean]: to save png-figure or print in command screen.
    """
    var2plot = {0: "dBZe", 1: "v", 2: "beta"}  # , 2: 'sigma' , 3: 'sigma', 4: 'kurt'
    # Dictionary for the vmax and vmin of the plot
    Vmin = {0: -55, 1: -1.5, 2: 0}  # , 2: 0 , 3: -3, 4: -3
    Vmax = {0: -20, 1: 1.5, 2: 10}  # , 2: 5 , 3: 3, 4: 3
    Vn = {0: 16, 1: 7, 2: 10}
    scale = {0: 1, 1: 1, 2: 1e6}
    titleStr = {
        0: "Reflectivity",
        1: "Vertical mean velocity",
        2: "Backscatter coeff.",
    }  # , 2: 'spectral width'
    cblabel = {
        0: "$Z_e, dBZe$",
        1: "$V_m, m/s$",
        2: r"$\beta$, $Mm^-1$",
    }  # , 2: 'spectral width'
    datestr = data["raw_time"][0].strftime("%Y%m%d")
    for idx in var2plot.keys():
        _var = var2plot[idx]
        # print(idx)
        print(_var)
        _fig, _axes = plt.subplots(nrows=1, figsize=(15, 6))
        _axes.set_facecolor("whitesmoke")
        cmap = mpl.cm.jet
        bounds = np.linspace(Vmin[idx], Vmax[idx], Vn[idx])
        #        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
        range_km = data["height"] / 1000.0
        q = _axes.pcolormesh(
            data["raw_time"],
            range_km,
            scale[idx] * data[_var].T,
            cmap=cmap,
            vmin=Vmin[idx],
            vmax=Vmax[idx],
        )
        q.cmap.set_over("white")
        q.cmap.set_under("darkblue")
        cb = plt.colorbar(q, ax=_axes, ticks=bounds, extend="max")
        cb.set_label(cblabel[idx])
        _axes.set_xlabel("Time, UTC")
        _axes.set_ylabel("Height, Km asl")
        _axes.set_title("%s on  %s" % (titleStr[idx], datestr))
        xmin = datetime.strptime(datestr, "%Y%m%d")
        xmax = xmin + timedelta(days=1)
        _axes.set_xlim(xmin, xmax)
        _axes.set_ylim(0, 2)
        if saveImageFlag:
            figstr = "%s_%s_%s.png" % ("cloudnet", var2plot[idx], datestr)
            finalpath = os.path.join(plt_conf["mainpath"], _var, figstr)
            print("Saving %s" % finalpath)
            final_dir_path = os.path.split(finalpath)[0]
            if not os.path.exists(final_dir_path):
                os.makedirs(final_dir_path)
            plt.savefig(finalpath, dpi=100, bbox_inches="tight")
            if os.path.exists(finalpath):
                print("Saving %s...DONE!" % finalpath)
            else:
                print("Saving %s... error!" % finalpath)
            plt.close(_fig)
        else:
            plt.show()


def plotLWP(data, mainpath, saveImageFlag):
    """
    Inputs:
    - data: from categorice_reader()
    - mainpath: "Y:\\datos\\CLOUDNET\\juelich\\quicklooks",
    - saveImageFlag [Boolean]: to save png-figure or print in command screen.
    """
    logger.warning("to be deprecated. Use cloudnetpy instead")
    datestr = data["raw_time"][0].strftime("%Y%m%d")
    fig, axes = plt.subplots(nrows=1, figsize=(15, 6))
    axes.set_facecolor("whitesmoke")
    plt.plot(data["raw_time"], data["lwp"])
    axes.set_xlabel("Time, UTC")
    axes.set_ylabel("LWP, $g/m^2$")
    axes.set_title("LWP on  %s" % datestr)
    xmin = datetime.strptime(datestr, "%Y%m%d")
    xmax = xmin + timedelta(days=1)
    axes.set_xlim(xmin, xmax)
    if saveImageFlag:
        figstr = "%s_%s_%s.png" % ("cloudnet", "lwp", datestr)
        finalpath = os.path.join(mainpath, "LWP", figstr)
        print("Saving %s" % finalpath)
        final_dir_path = os.path.split(finalpath)[0]
        if not os.path.exists(final_dir_path):
            os.makedirs(final_dir_path)
        plt.savefig(finalpath, dpi=100, bbox_inches="tight")
        if os.path.exists(finalpath):
            print("Saving %s...DONE!" % finalpath)
        else:
            print("Saving %s... error!" % finalpath)
        plt.close(fig)
    else:
        plt.show()



from pathlib import Path
import numpy as np
import xarray as xr
from rpgpy import read_rpg
from rpgpy.spcutil import scale_spectra

# from gfatpy.radar._dealiazing import dealiaze, dealiazeOneHeight


class rpg:
    def __init__(self, path: Path):
        self.path = path
        self.type = path.name.split(".")[0].split("_")[-1]
        self.level = int(path.name.split(".")[-1][-1])
        self._raw = None
        self._header = None
        self._band = None
        self._dataset = None

    @property
    def raw(self) -> dict:
        if self._raw is None:
            _, self._raw = read_rpg(self.path)
        return self._raw

    @property
    def header(self) -> dict:
        if self._header is None:
            self._header, _ = read_rpg(self.path)
        return self._header

    @property
    def band(self) -> str:
        if self._band is None:
            if self.header["Freq"] > 75:
                self._band = "W"
            else:
                self._band = "Ka"
        return self._band

    @property
    def dataset(
        self, time_parser: str = "seconds since 2001-01-01 00:00:00 UTC"
    ) -> xr.Dataset:
        if self._dataset == None:
            coords = {
                "time": self.raw["Time"],
                "range": self.header["RAlts"],
                "spectrum": np.arange(self.header["velocity_vectors"].shape[1]),
                "chirp": np.arange(self.header["velocity_vectors"].shape[0]),
            }
            dataset = xr.Dataset(coords=coords)
            dataset["time"].attrs = {"long_name": "time", "units": time_parser}
            dataset = xr.decode_cf(dataset)

            dataset["chirp_number"] = xr.DataArray(
                np.zeros(self.header["RAlts"].size),
                coords={"range": self.header["RAlts"]},
            )
            for idx, start_chirp_ in enumerate(self.header["RngOffs"]):
                dataset["chirp_number"].loc[
                    {"range": dataset.range[start_chirp_:]}
                ] = idx
            dataset = dataset.assign_coords(
                chirp=np.sort(np.unique(dataset["chirp_number"].values.astype(int)))
            )

            dataset['nyquist_velocity'] = (
                ("chirp"),
                self.header["MaxVel"],
            )
            dataset['maximum_range'] = (
                ("chirp"),
                self.header["RangeMax"],
            )
            dataset['minimum_range'] = (
                ("chirp"),
                self.header["RangeMax"],
            )
            dataset['doppler_spectrum_length'] = (
                ("chirp"),
                self.header['SpecN'],
            )
            
            dataset["velocity_vectors"] = (
                ("chirp", "spectrum"),
                self.header["velocity_vectors"],
            )
            dataset["doppler_spectrum_h"] = (
                ("time", "range", "spectrum"),
                self.raw["HSpec"],
            )
            dataset["covariance_spectrum_re"] = (
                ("time", "range", "spectrum"),
                self.raw["ReVHSpec"],
            )
            dataset["doppler_spectrum"] = (
                ("time", "range", "spectrum"),
                scale_spectra(self.raw["TotSpec"], self.header["SWVersion"]),
            )
            dataset["doppler_spectrum_v"] = xr.DataArray(
            dims=dataset["doppler_spectrum"].dims,
            data=dataset["doppler_spectrum"].values
            - dataset["doppler_spectrum_h"].values
            - 2 * dataset["covariance_spectrum_re"].values,
            )
            dataset["sZDR"] = xr.DataArray(
                dims=dataset["doppler_spectrum_h"].dims,
                data=10 * np.log10(dataset["doppler_spectrum_h"].values)
                - 10 * np.log10(dataset["doppler_spectrum_v"].values),
            )
            dataset["sZDRmax"] = dataset["sZDR"].max(dim="spectrum")
            self._dataset = dataset
        return self._dataset

    # def dealiaze(self, dataset, target_ranges= float | tuple[float, float])-> xr.Dataset:
    #     if isinstance( target_ranges, float):
    #         dataset = dealiazeOneHeight(target_ranges, dataset)
    #     else:
    #         dealiaze(dataset)
    #     return 


general_parameters = \
    {'System': '\'VELETA\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
     'Latitude_degrees_north': 37.164,
     'Longitude_degrees_east': -3.605,
     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

channel_parameters = \
    {'BT0': {'channel_ID': 1363,                    
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'Laser_Shots': 600,
                    'LR_Input': 1,
                    'DAQ_Range': 500.0, },
     'BT1': {'channel_ID': 315,                    
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'Laser_Shots': 600,
                    'LR_Input': 1,
                    'DAQ_Range': 500.0, },
     }


from pdb import set_trace
from matplotlib.patches import Circle
import numpy as np


def circular_grid(
    ax,
    radius: list | np.ndarray,
    rlabels: bool = True,
    label_angle: float = 45,
    relative_offset: tuple[float, float] = (0.25, 0.25),
    rlabel_fmt: str = "%.1f",
    color: str = "black",
    linewidth: float = 0.5,
    linestyle: str = "--",
):
    if isinstance(radius, list):
        radius = np.array(radius)
    elif not isinstance(radius, np.ndarray):
        raise TypeError("radius must be a list or numpy array.")

    #Leave only positive values and should be unique
    radius = np.unique(radius[radius > 0])

    # Add circular grid lines
    for r in radius:  # Specify the radii for the circles
        circle = Circle((0, 0), r, color=color, linewidth=linewidth, linestyle=linestyle, fill=False)
        ax.add_artist(circle)
    
    if rlabels:
        for r in radius:
            label_x = r * np.cos(np.deg2rad(label_angle)) + relative_offset[0]
            label_y = r * np.sin(np.deg2rad(label_angle)) + relative_offset[1]
            ax.text(
                    label_x, label_y,
                    rlabel_fmt % r,
                    verticalalignment="center",
                    horizontalalignment="center",
                )


from datetime import datetime
import re
import sys
from pathlib import Path
import atexit

from linc.config import get_config
from loguru import logger

from gfatpy.utils.io import find_nearest_filepath
from .types import Measurement

from .utils import (
    cleanup_tmp_folders,
    filter_by_type,
    launch_write_nc,
    to_measurements,
)
from gfatpy.lidar.utils.types import LidarName, MeasurementType, Telescope


CONFIGS_DIR = Path(__file__).parent / "configs"

logger.add(sys.stdout, level="INFO")

# Register cleanup function to be executed when the interpreter exits
atexit.register(cleanup_tmp_folders)


def measurements_to_nc(
    measurements: list[Measurement],
    lidar_name: LidarName,
    telescope: Telescope = Telescope.xf,
    raw_dir: Path | str | None = None,
    output_dir: Path | str | None = None,
    config_file: Path | str | None = None,
    number_of_previous_days: int = 5,
) -> None:
    """Converts raw data to netcdf files.

    Args:

        - measurements (list[Measurement]): Measurements to convert.
        - lidar_name (LidarName): Name of the lidar.
        - telescope (Telescope, optional): Telescope used. Defaults to Telescope.xf.
        - raw_dir (Path | str | None, optional): Raw data directory where previous data will be searched. Defaults to None means no previous data are searched.
        - output_dir (Path | str | None, optional): Output directory. Defaults to None.
        - config_file (Path | str | None, optional): Configuration file. Defaults to None.
        - number_of_previous_days (int, optional): Number of previous days to search. Defaults to 5.
        - convert_dc (bool, optional): Force conversion of ALL DC measurements. Defaults to False means only DC linked to other files will be converted.

    Raises:

        - ValueError: Value of output_dir is not a directory.
        - ValueError: Value of raw_dir is not a directory.
        - Exception: Linked DC measurement to RS|HF not found.
        - Exception: Linked DC measurement to DP|TC|OT not found.
    """

    rs_measurements = filter_by_type(measurements, MeasurementType.RS)
    hf_measurements = filter_by_type(measurements, MeasurementType.HF)
    dc_measurements = filter_by_type(measurements, MeasurementType.DC)
    dp_measurements = filter_by_type(measurements, MeasurementType.DP)
    tc_measurements = filter_by_type(measurements, MeasurementType.TC)
    ot_measurements = filter_by_type(measurements, MeasurementType.OT)

    if len(rs_measurements) != 0:
        rs_hf_measurements_to_nc(
            rs_measurements,
            lidar_name,
            telescope,
            raw_dir,
            output_dir,
            config_file,
            number_of_previous_days,
            measurement_type=MeasurementType.RS,
        )
    if len(hf_measurements) != 0:
        rs_hf_measurements_to_nc(
            hf_measurements,
            lidar_name,
            telescope,
            raw_dir,
            output_dir,
            config_file,
            number_of_previous_days,
            measurement_type=MeasurementType.HF,
        )
    if len(dc_measurements) != 0:
        dc_measurements_to_nc(
            dc_measurements,
            lidar_name,
            telescope,
            raw_dir,
            output_dir,
            config_file,
        )
    if len(dp_measurements) != 0:
        dp_tc_ot_measurements_to_nc(
            dp_measurements,
            lidar_name,
            telescope,
            raw_dir,
            output_dir,
            config_file,
            measurement_type=MeasurementType.DP,
        )
    if len(tc_measurements) != 0:
        dp_tc_ot_measurements_to_nc(
            tc_measurements,
            lidar_name,
            telescope,
            raw_dir,
            output_dir,
            config_file,
            measurement_type=MeasurementType.TC,
        )
    if len(ot_measurements) != 0:
        dp_tc_ot_measurements_to_nc(
            ot_measurements,
            lidar_name,
            telescope,
            raw_dir,
            output_dir,
            config_file,
            measurement_type=MeasurementType.OT,
        )


def rs_hf_measurements_to_nc(
    measurements,
    lidar_name: LidarName,
    telescope: Telescope = Telescope.xf,
    raw_dir: Path | str | None = None,
    output_dir: Path | str | None = None,
    config_file: Path | str | None = None,
    number_of_previous_days: int = 5,
    measurement_type: MeasurementType = MeasurementType.RS,
) -> None:
    """Converts raw data to netcdf files.

    Args:

        - measurements (list[Measurement]): Measurements to convert.
        - lidar_name (LidarName): Name of the lidar.
        - telescope (Telescope, optional): Telescope used. Defaults to Telescope.xf.
        - raw_dir (Path | str | None, optional): Raw data directory where previous data will be searched. Defaults to None means no previous data are searched.
        - output_dir (Path | str | None, optional): Output directory. Defaults to None.
        - config_file (Path | str | None, optional): Configuration file. Defaults to None.
        - number_of_previous_days (int, optional): Number of previous days to search. Defaults to 5.
        - convert_dc (bool, optional): Force conversion of ALL DC measurements. Defaults to False means only DC linked to other files will be converted.

    Raises:

        - ValueError: Value of output_dir is not a directory.
        - ValueError: Value of raw_dir is not a directory.
        - Exception: Linked DC measurement to RS|HF not found.
        - Exception: Linked DC measurement to DP|TC|OT not found.
    """

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError(f"{output_dir} not found.")

    # Get lidar configuration file
    target_datetime = measurements[0].datetime
    config_filepath = search_config_file(lidar_name, target_datetime, config_file)
    config = get_config(config_filepath)

    search_previous_days = True

    # Check directory where data of previous days should be stored
    if isinstance(raw_dir, str):
        raw_dir = Path(raw_dir)
        if not raw_dir.exists() or not raw_dir.is_dir():
            raise ValueError(f"{raw_dir} not found.")
    elif raw_dir is None:
        search_previous_days = False

    measurements = filter_by_type(measurements, measurement_type)

    if len(measurements) == 0:
        return None
    # get the first measurement:
    measurement = measurements[0]

    # Add measurements of the same type and date to the measurement object
    measurement.linked_measurements = [
        m
        for m in measurements
        if m.path != measurement.path
        and m.type == measurement_type
        and m.datetime.date() == measurement.datetime.date()
    ]

    # Add previous days to the search
    if search_previous_days:
        prev_paths = measurement.find_prev_paths(
            number_of_previous_days, lidar_name, raw_dir  # type: ignore
        )
        for prev_path in prev_paths:
            prev_measurements = to_measurements(prev_path.glob("[RH][SF]*"))
            if len(prev_measurements) != 0:
                linked_ = [
                    prev_
                    for prev_ in prev_measurements
                    if prev_.has_target_date(measurement.datetime)
                ]
                measurement.add_linked_measurements(linked_)

    signal_type = measurement_type
    try:
        files2convert = measurement.get_filepaths()
    except Exception as e:
        raise RuntimeError(f"Error in {measurement.path}: {e}")
    if len(files2convert) != 0:
        launch_write_nc(
            measurement,
            files2convert,
            output_dir,
            lidar_name,
            telescope,
            signal_type,
            config,
        )
    return None


def dc_measurements_to_nc(
    measurements,
    lidar_name: LidarName,
    telescope: Telescope = Telescope.xf,
    raw_dir: Path | str | None = None,
    output_dir: Path | str | None = None,
    config_file: Path | str | None = None,
) -> None:
    """Converts raw data to netcdf files.

    Args:

        - measurements (list[Measurement]): Measurements to convert.
        - lidar_name (LidarName): Name of the lidar.
        - telescope (Telescope, optional): Telescope used. Defaults to Telescope.xf.
        - raw_dir (Path | str | None, optional): Raw data directory where previous data will be searched. Defaults to None means no previous data are searched.
        - output_dir (Path | str | None, optional): Output directory. Defaults to None.
        - config_file (Path | str | None, optional): Configuration file. Defaults to None.

    Raises:

        - ValueError: Value of output_dir is not a directory.
        - ValueError: Value of raw_dir is not a directory.
        - Exception: Linked DC measurement to RS|HF not found.
        - Exception: Linked DC measurement to DP|TC|OT not found.
    """

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError(f"{output_dir} not found.")

    # Get lidar configuration file
    target_datetime = measurements[0].datetime
    config_filepath = search_config_file(lidar_name, target_datetime, config_file)
    config = get_config(config_filepath)

    # Check directory where data of previous days should be stored
    if isinstance(raw_dir, str):
        raw_dir = Path(raw_dir)
        if not raw_dir.exists() or not raw_dir.is_dir():
            raise ValueError(f"{raw_dir} not found.")

    measurements = filter_by_type(measurements, MeasurementType.DC)

    if len(measurements) == 0:
        return None
    for measurement in measurements:
        # Gather files to convert
        files2convert = measurement.get_filepaths()
        if len(files2convert) != 0:
            launch_write_nc(
                measurement,
                files2convert,
                output_dir,
                lidar_name,
                telescope,
                signal_type=MeasurementType.RS,
                config=config,
            )
    return None


def dp_tc_ot_measurements_to_nc(
    measurements,
    lidar_name: LidarName,
    telescope: Telescope = Telescope.xf,
    raw_dir: Path | str | None = None,
    output_dir: Path | str | None = None,
    config_file: Path | str | None = None,
    measurement_type: MeasurementType = MeasurementType.DP,
) -> None:
    """Converts raw data to netcdf files.

    Args:

        - measurements (list[Measurement]): Measurements to convert.
        - lidar_name (LidarName): Name of the lidar.
        - telescope (Telescope, optional): Telescope used. Defaults to Telescope.xf.
        - raw_dir (Path | str | None, optional): Raw data directory where previous data will be searched. Defaults to None means no previous data are searched.
        - output_dir (Path | str | None, optional): Output directory. Defaults to None.
        - config_file (Path | str | None, optional): Configuration file. Defaults to None.
        - number_of_previous_days (int, optional): Number of previous days to search. Defaults to 5.
        - convert_dc (bool, optional): Force conversion of ALL DC measurements. Defaults to False means only DC linked to other files will be converted.

    Raises:

        - ValueError: Value of output_dir is not a directory.
        - ValueError: Value of raw_dir is not a directory.
        - Exception: Linked DC measurement to RS|HF not found.
        - Exception: Linked DC measurement to DP|TC|OT not found.
    """

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    elif output_dir is None:
        output_dir = Path.cwd()

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError(f"{output_dir} not found.")

    # Get lidar configuration file
    target_datetime = measurements[0].datetime
    config_filepath = search_config_file(lidar_name, target_datetime, config_file)
    config = get_config(config_filepath)

    # Check directory where data of previous days should be stored
    if isinstance(raw_dir, str):
        raw_dir = Path(raw_dir)
        if not raw_dir.exists() or not raw_dir.is_dir():
            raise ValueError(f"{raw_dir} not found.")

    measurements = filter_by_type(measurements, measurement_type)
    if len(measurements) == 0:
        return None

    for measurement in measurements:
        # files2convert is taken for each subfolder
        for subdir in measurement.sub_dirs:
            if subdir == measurement.datetime.strftime("%Y%m%d"):
                continue
            try:
                escaped_subdir = re.escape(subdir)
                files2convert = measurement.get_filepaths(
                    pattern_or_list=rf".*{escaped_subdir}.*"
                )
            except:
                raise RuntimeError(f"Error in {measurement.path}")
            if len(files2convert) != 0:
                launch_write_nc(
                    measurement,
                    files2convert,
                    output_dir,
                    lidar_name,
                    telescope,
                    subdir,
                    config,
                )
    return None


def search_config_file(
    lidar_name: LidarName,
    target_datetime: datetime,
    opt_config: Path | str | None,
) -> Path:
    """Searches for a configuration file.

    Args:
        lidar_name (LidarName): Name of the lidar.
        opt_config (Path | str | None, optional): Path to the configuration file. Defaults to None.

    Raises:
        FileNotFoundError: No configution file found in opt_config.

    Returns:
        Config: Configuration object.
    """
    if opt_config is not None:
        if isinstance(opt_config, Path):
            config_path = opt_config
        elif isinstance(opt_config, str):
            config_path = Path(opt_config)

        if not config_path.exists():
            raise FileNotFoundError(f"Configution file {opt_config} not found.")
    else:
        config_dir = Path(__file__).parent / "configs"
        config_path = find_nearest_filepath(
            config_dir,
            f"{lidar_name.value}*.toml",
            1,
            target_datetime,
            and_previous=True,
        )
        if not config_path.exists():
            raise FileNotFoundError(f"No configution file found in {config_dir}")
    return config_path
